# This code is part of Qiskit.
#
# (C) Copyright IBM 2022.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.
"""
Circuit basis for tomography preparation and measurement circuits
"""
from typing import Sequence, Optional, Tuple, Union, List, Dict
import numpy as np
from qiskit.circuit import QuantumCircuit, Instruction
from qiskit.quantum_info.states.quantum_state import QuantumState
from qiskit.quantum_info.operators.base_operator import BaseOperator
from qiskit.quantum_info.operators.channel.quantum_channel import QuantumChannel
from qiskit.quantum_info import DensityMatrix, Statevector, Operator, SuperOp
from qiskit.exceptions import QiskitError

from .base_basis import PreparationBasis, MeasurementBasis
from .cache_method import cache_method, _method_cache_name


# Typing
Povm = Union[List[Statevector], List[DensityMatrix], QuantumChannel]
States = Union[List[QuantumState], Dict[Tuple[int, ...], QuantumState]]


class LocalPreparationBasis(PreparationBasis):
    """Local tensor-product preparation basis.

    This basis consists of a set of 1-qubit instructions which
    are used to define a tensor-product basis on N-qubits.
    """

    def __init__(
        self,
        name: str,
        instructions: Optional[Sequence[Instruction]] = None,
        default_states: Optional[States] = None,
        qubit_states: Optional[Dict[Tuple[int, ...], States]] = None,
    ):
        """Initialize a fitter preparation basis.

        Args:
            name: a name to identify the basis.
            instructions: list of 1-qubit instructions for preparing states
                          from the :math:`|0\\rangle` state.
            default_states: Optional, default density matrices prepared by the
                            input instructions. If None these will be determined by
                            ideal simulation of the preparation instructions.
            qubit_states: Optional, a dict with physical qubit keys and a list of
                          density matrices prepared by the list of basis instructions
                          for a specific qubit. The default states will be used for any
                          qubits not specified in this dict.

        Raises:
            QiskitError: If input states or instructions are not valid, or no
                         instructions or states are provided.
        """
        if instructions is None and default_states is None and qubit_states is None:
            raise QiskitError(
                "LocalPreparationBasis must define at least one of instructions, "
                "default_states, or qubit_states."
            )
        super().__init__(name)
        # POVM element variables
        self._instructions = _format_instructions(instructions)
        self._default_states = _format_states(default_states, (0,), self._instructions)
        self._qubit_states = _format_qubit_states(qubit_states)
        self._custom_defaults = bool(default_states)

        # Other attributes derived from povms and instructions
        # that need initializing
        self._qubits = set()
        self._size = None
        self._default_dim = None
        self._qubit_dim = {}
        self._hash = None

        # Initialize attributes
        self._initialize()

    def __repr__(self):
        return f"<{type(self).__name__}: {self.name}>"

    def __hash__(self):
        return self._hash

    def __eq__(self, value):
        return (
            super().__eq__(value)
            and self._size == getattr(value, "_size", None)
            and self._default_dim == getattr(value, "_default_dim", None)
            and self._custom_defaults == getattr(value, "_custom_defaults", None)
            and self._qubits == getattr(value, "_qubits", None)
            and self._qubit_dim == getattr(value, "_qubit_dim", None)
            and self._instructions == getattr(value, "_instructions", None)
        )

    def index_shape(self, qubits: Sequence[int]) -> Tuple[int, ...]:
        return len(qubits) * (self._size,)

    def matrix_shape(self, qubits: Sequence[int]) -> Tuple[int, ...]:
        qubits = tuple(qubits)
        if len(qubits) > 1 and qubits in self._qubit_dim:
            return self._qubit_dim[qubits]
        dims = tuple()
        for i in qubits:
            dims += self._qubit_dim.get((i,), (self._default_dim,))
        return dims

    def circuit(
        self, index: Sequence[int], qubits: Optional[Sequence[int]] = None
    ) -> QuantumCircuit:
        # pylint: disable = unused-argument
        if not self._instructions:
            raise NotImplementedError(
                f"Basis {self.name} does not define circuits so can only be "
                " used as a fitter basis for analysis."
            )
        return _tensor_product_circuit(self._instructions, index, self._name)

    def matrix(self, index: Sequence[int], qubits: Optional[Sequence[int]] = None):
        # Convert args to hashable tuples
        if qubits is None:
            qubits = tuple(range(len(index)))
        else:
            qubits = tuple(qubits)
        index = tuple(index)

        try:
            # Look for custom POVM for specified qubits
            state = self._generate_qubits_state(index, qubits)
            if state is not None:
                return state.data

            mat = np.eye(1)
            for idx, qubit in zip(index, qubits):
                qubit_state = self._generate_qubits_state((idx,), (qubit,))
                mat = np.kron(qubit_state, mat)
            return mat

        except TypeError as ex:
            # This occurs if basis is constructed with qubit_states
            # kwarg but no default_states or instructions and is called for
            # a qubit not in the specified kwargs.
            raise ValueError(f"Invalid qubits for basis {self.name}") from ex

    @cache_method()
    def _generate_qubits_state(self, index: Tuple[int, ...], qubits: Tuple[int, ...]):
        """LRU cached function for returning POVMS"""
        num_qubits = len(qubits)

        # Check for N-qubit states
        if qubits in self._qubit_states:
            # Get states for specified qubits
            # TODO: In the future we could add support for different orderings
            #       of qubits by permuting the returned POVMS
            states = self._qubit_states[qubits]
            if index in states:
                return states[index]

            # Look up custom 0 init state for specified qubits
            # TODO: Add support for noisy instructions
            if not self._instructions:
                raise NotImplementedError(
                    f"Basis {self.name} does not define circuits to construct POVMs from"
                )
            key0 = num_qubits * (0,)
            if key0 in states:
                circuit = _tensor_product_circuit(self._instructions, index, self._name)
                return _generate_state(circuit, states[key0])

        # No match, so if 1-qubit use default, otherwise return None
        if num_qubits == 1 and self._default_states:
            return self._default_states[index]
        return None

    def _initialize(self):
        """Initialize dimension and num outcomes"""
        if self._instructions:
            self._size = len(self._instructions)

        # Format default POVMs
        if self._default_states:
            default_state = next(iter(self._default_states.values()))
            self._default_dim = np.prod(default_state.dims())
            if self._size is None:
                self._size = len(self._default_states)
            elif len(self._default_states) != self._size:
                raise QiskitError("Number of instructions and number of states must be equal.")

        # Format qubit states
        for qubits, states in self._qubit_states.items():
            state = next(iter(states.values()))
            num_qubits = len(qubits)
            dims = state.dims()
            if num_qubits == 1:
                qubit_dim = (np.prod(dims),)
            elif len(dims) == num_qubits:
                qubit_dim = tuple(dims)
            else:
                # Assume all subsystems have the same dimension if the provided
                # state dimension don't match number of qubits
                ave_dim = np.prod(dims) ** (1 / num_qubits)
                if int(ave_dim) != ave_dim:
                    raise QiskitError("Cannot infer unequal subsystem dimensions from input states")
                qubit_dim = num_qubits * (int(ave_dim),)
            self._qubit_dim[qubits] = qubit_dim
            self._qubits.update(qubits)

        # Pseudo hash value to make basis hashable for LRU cached functions
        self._hash = hash(
            (
                type(self),
                self._name,
                self._size,
                self._default_dim,
                self._custom_defaults,
                tuple(self._qubits),
                tuple(sorted(self._qubit_dim.items())),
                tuple(type(i) for i in self._instructions),
            )
        )

    def __json_encode__(self):
        value = {
            "name": self._name,
            "instructions": list(self._instructions) if self._instructions else None,
        }
        if self._custom_defaults:
            value["default_states"] = self._default_states
        if self._qubit_states:
            value["qubit_states"] = self._qubit_states
        return value

    def __getstate__(self):
        # override get state to skip class cache when pickling
        state = self.__dict__.copy()
        state.pop(_method_cache_name(self), None)
        return state


class LocalMeasurementBasis(MeasurementBasis):
    """Local tensor-product measurement basis.

    This basis consists of a set of 1-qubit instructions which
    are used to define a tensor-product basis on N-qubits to
    rotate a desired multi-qubit measurement basis to the Z-basis
    measurement.
    """

    def __init__(
        self,
        name: str,
        instructions: Optional[Sequence[Instruction]] = None,
        default_povms: Optional[Sequence[Povm]] = None,
        qubit_povms: Optional[Dict[Tuple[int, ...], Sequence[Povm]]] = None,
    ):
        """Initialize a fitter preparation basis.

        Args:
            name: a name to identity the basis.
            instructions: list of instructions for rotating a desired
                          measurement basis to the standard :math:`Z^{\\otimes n}`
                          computational basis measurement.
            default_povms: Optional, list if positive operators valued measures (POVM)
                           for of the measurement basis instructions. A POVM can be
                           input as a list of effects (Statevector or DensityMatrix)
                           for each possible measurement outcome of that basis, or as
                           a single QuantumChannel. For the channel case the effects
                           will be calculated by evolving the computation basis states
                           by the adjoint of the channel. If None the input instructions
                           will be used as the POVM channel.
            qubit_povms: Optional, a dict with physical qubit keys and a list of POVMs
                         corresponding to each basis measurement instruction for the
                         specific qubit. The default POVMs will be used for any qubits
                         not specified in this dict.

        Raises:
            QiskitError: If the input instructions or POVMs are not valid, or if no
                         instructions or POVMs are provided.
        """
        if instructions is None and default_povms is None and qubit_povms is None:
            raise QiskitError(
                "LocalMeasurementBasis must define at least one of instructions, "
                "default_povms, or qubit_povms."
            )
        super().__init__(name)

        # POVM element variables
        self._instructions = _format_instructions(instructions)
        self._default_povms = _format_default_povms(default_povms, self._instructions)
        self._qubit_povms = _format_qubit_povms(qubit_povms)
        self._custom_defaults = bool(default_povms)

        # Other attributes derived from povms and instructions
        # that need initializing
        self._qubits = set()
        self._size = None
        self._default_num_outcomes = None
        self._default_dim = None
        self._qubit_num_outcomes = {}
        self._qubit_dim = {}
        self._hash = None

        # Initialize attributes
        self._initialize()

    def __repr__(self):
        return f"<{type(self).__name__}: {self.name}>"

    def __hash__(self):
        return self._hash

    def __eq__(self, value):
        return (
            super().__eq__(value)
            and self._size == getattr(value, "_size", None)
            and self._default_dim == getattr(value, "_default_dim", None)
            and self._default_num_outcomes == getattr(value, "_default_num_outcomes", None)
            and self._custom_defaults == getattr(value, "_custom_defaults", None)
            and self._qubit_dim == getattr(value, "_qubit_dim", None)
            and self._qubit_num_outcomes == getattr(value, "_qubit_num_outcomes", None)
            and self._qubits == getattr(value, "_qubits", None)
            and self._instructions == getattr(value, "_instructions", None)
        )

    def index_shape(self, qubits: Sequence[int]) -> Tuple[int, ...]:
        return len(qubits) * (self._size,)

    def matrix_shape(self, qubits: Sequence[int]) -> Tuple[int, ...]:
        qubits = tuple(qubits)
        if len(qubits) > 1 and qubits in self._qubit_dim:
            return self._qubit_dim[qubits]
        dims = tuple()
        for i in qubits:
            dims += self._qubit_dim.get((i,), (self._default_dim,))
        return dims

    def outcome_shape(self, qubits: Sequence[int]) -> Tuple[int, ...]:
        qubits = tuple(qubits)
        if qubits in self._qubit_num_outcomes:
            return self._qubit_num_outcomes[qubits]
        shape = tuple()
        for i in qubits:
            shape += self._qubit_num_outcomes.get((i,), (self._default_num_outcomes,))
        return shape

    def circuit(self, index: Sequence[int], qubits: Optional[Sequence[int]] = None):
        # pylint: disable = unused-argument
        if not self._instructions:
            raise NotImplementedError(
                f"Basis {self.name} does not define circuits so can only be "
                " used as a fitter basis for analysis."
            )
        circuit = _tensor_product_circuit(self._instructions, index, self._name)
        circuit.measure_all()
        return circuit

    def matrix(self, index: Sequence[int], outcome: int, qubits: Optional[Sequence[int]] = None):
        # Convert args to hashable tuples
        if qubits is None:
            qubits = tuple(range(len(index)))
        else:
            qubits = tuple(qubits)
        index = tuple(index)

        try:
            # Look for custom POVM for specified qubits
            qubit_povm = self._generate_qubits_povm(index, qubits)
            if qubit_povm:
                return qubit_povm[outcome].data

            # Otherwise construct tensor product POVM
            outcome_index = self._outcome_indices(outcome, qubits)
            mat = np.eye(1)
            for idx, odx, qubit in zip(index, outcome_index, qubits):
                povm = self._generate_qubits_povm((idx,), (qubit,))
                mat = np.kron(povm[odx], mat)
            return mat

        except TypeError as ex:
            # This occurs if basis is constructed with qubit_states
            # kwarg but no default_states or instructions and is called for
            # a qubit not in the specified kwargs.
            raise ValueError(f"Invalid qubits for basis {self.name}") from ex

    def _initialize(self):
        """Initialize dimension and num outcomes"""
        if self._instructions:
            self._size = len(self._instructions)

        # Format default POVMs
        if self._default_povms:
            default_povm = next(iter(self._default_povms.values()))
            self._default_num_outcomes = len(default_povm)
            self._default_dim = np.prod(default_povm[0].dims())
            if self._size is None:
                self._size = len(self._default_povms)
            elif len(self._default_povms) != self._size:
                raise QiskitError("Number of instructions and number of states must be equal.")
            if any(
                len(povm) != self._default_num_outcomes for povm in self._default_povms.values()
            ):
                raise QiskitError(
                    "LocalMeasurementBasis default POVM elements must all have "
                    "the same number of outcomes."
                )

        # Format qubit POVMS
        for qubits, povms in self._qubit_povms.items():
            povm = next(iter(povms.values()))
            num_povms = len(povm)
            if any(len(povm) != num_povms for povm in povms.values()):
                raise QiskitError(
                    "LocalMeasurementBasis POVM elements must all have the "
                    "same number of outcomes."
                )
            num_qubits = len(qubits)
            dims = povm[0].dims()
            dim = np.prod(dims)
            if num_qubits == 1:
                qubit_dim = (dim,)
                num_outcomes = (num_povms,)
            elif len(dims) == num_qubits:
                qubit_dim = tuple(dims)
                if dim != num_povms:
                    raise QiskitError("POVMs dimensions don't match number of outcomes")
                num_outcomes = qubit_dim
            else:
                # Assume all subsystems have the same dimension if the provided
                # operator dimension don't match number of qubits
                ave_dim = np.prod(dims) ** (1 / num_qubits)
                if int(ave_dim) != ave_dim:
                    raise QiskitError("Cannot infer unequal subsystem dimensions from input POVMs")
                qubit_dim = num_qubits * (int(ave_dim),)
                ave_num_outcomes = num_povms ** (1 / num_qubits)
                if int(ave_num_outcomes) != ave_num_outcomes:
                    raise QiskitError("Cannot infer unequal subsystem num_outcome from input POVMs")
                num_outcomes = num_qubits * (int(ave_dim),)
            self._qubit_num_outcomes[qubits] = num_outcomes
            self._qubit_dim[qubits] = qubit_dim
            self._qubits.update(qubits)

        # Pseudo hash value to make basis hashable for LRU cached functions
        self._hash = hash(
            (
                type(self),
                self._name,
                self._size,
                self._default_dim,
                self._default_num_outcomes,
                self._custom_defaults,
                tuple(self._qubits),
                tuple(sorted(self._qubit_dim.items())),
                tuple(sorted(self._qubit_num_outcomes.items())),
                tuple(type(i) for i in self._instructions),
            )
        )

    @cache_method()
    def _outcome_indices(self, outcome: int, qubits: Tuple[int, ...]) -> Tuple[int, ...]:
        """Convert an outcome integer to a tuple of single-qubit outcomes"""
        num_outcomes = np.prod(self._qubit_num_outcomes.get(qubits[:1], self._default_num_outcomes))
        try:
            value = (outcome % num_outcomes,)
            if len(qubits) == 1:
                return value
            return value + self._outcome_indices(outcome // num_outcomes, qubits[1:])
        except TypeError as ex:
            raise ValueError("Invalid qubits for basis") from ex

    @cache_method()
    def _generate_qubits_povm(self, index: Tuple[int, ...], qubits: Tuple[int, ...]):
        """LRU cached function for returning POVMS"""
        num_qubits = len(qubits)
        if qubits in self._qubit_povms:
            # Get POVMS for specified qubits
            # TODO: In the future we could add support for different orderings
            #       of qubits by permuting the returned POVMS
            povms = self._qubit_povms[qubits]
            if index in povms:
                return povms[index]

            # Look up custom Z-default POVM for specified qubits
            # TODO: Add support for noisy instructions
            if not self._instructions:
                raise NotImplementedError(
                    f"Basis {self.name} does not define circuits to construct POVMs from"
                )
            key0 = num_qubits * (0,)
            if key0 in povms:
                circuit = _tensor_product_circuit(self._instructions, index, self._name)
                return _generate_povm(circuit, povms[key0])

        # No match, so if 1-qubit use default, otherwise return None
        if num_qubits == 1 and self._default_povms:
            return self._default_povms[index[0]]
        return None

    def __json_encode__(self):
        value = {
            "name": self._name,
            "instructions": self._instructions if self._instructions else None,
        }
        if self._custom_defaults:
            value["default_povms"] = self._default_povms
        if self._qubit_povms:
            value["qubit_povms"] = self._qubit_povms
        return value

    def __getstate__(self):
        # override get state to skip class cache when pickling
        state = self.__dict__.copy()
        state.pop(_method_cache_name(self), None)
        return state


def _tensor_product_circuit(
    instructions: Sequence[Instruction],
    index: Sequence[int],
    name: str = "",
) -> QuantumCircuit:
    """Return tensor product of 1-qubit basis instructions"""
    size = len(instructions)
    circuit = QuantumCircuit(len(index), name=f"{name}{list(index)}")
    for i, elt in enumerate(index):
        if elt >= size:
            raise QiskitError("Invalid basis element index")
        circuit.append(instructions[elt], [i])
    return circuit


def _format_instructions(instructions: Sequence[any]) -> List[Instruction]:
    """Parse multiple input formats for list of instructions"""
    ret = []
    if instructions is None:
        return ret
    for inst in instructions:
        # Convert to instructions if object is not an instruction
        # This allows converting raw unitary matrices and other operator
        # types like Pauli or Clifford into instructions.
        if not isinstance(inst, Instruction):
            if hasattr(inst, "to_instruction"):
                inst = inst.to_instruction()
            else:
                inst = Operator(inst).to_instruction()

        # Validate that instructions are single qubit
        if inst.num_qubits != 1:
            raise QiskitError(f"Input instruction {inst.name} is not a 1-qubit instruction.")
        ret.append(inst)

    return ret


def _generate_povm(
    value: Union[List[DensityMatrix], Instruction, Operator, SuperOp],
    default_z: Optional[List[DensityMatrix]] = None,
    dims: Optional[Tuple[int, ...]] = None,
) -> List[DensityMatrix]:
    """Format a POVM into list of density matrix effects"""
    # If already a list convert to DensityMatrix objects
    if isinstance(value, (list, tuple)):
        return [DensityMatrix(i, dims=dims) for i in value]

    # Otherwise convert from operator/channel to POVM effects
    try:
        chan = Operator(value)
    except QiskitError:
        chan = SuperOp(value)
    adjoint = chan.adjoint()
    if dims is None:
        dims = adjoint.input_dims()

    if default_z is not None:
        z_states = [DensityMatrix(i, dims) for i in default_z]
    else:
        z_states = [DensityMatrix.from_int(i, dims) for i in range(np.prod(dims))]

    return [state.evolve(adjoint) for state in z_states]


def _format_default_povms(
    default_povms: any, instructions: Optional[Sequence[Instruction]] = None
) -> Dict[Tuple[int, ...], List[DensityMatrix]]:
    "Format default POVM data"
    # Parse data into a dict
    # Legacy data handling
    if isinstance(default_povms, (list, tuple, np.ndarray)):
        povms = dict(enumerate(default_povms))
    elif isinstance(default_povms, dict):
        povms = {int(key): val for key, val in default_povms.items()}
    elif not default_povms:
        povms = {}

    # Add instructions to POVM dict if not specified in data
    if instructions and len(povms) < len(instructions):
        for i, inst in enumerate(instructions):
            if i not in povms:
                povms[i] = inst

    # Look for default POVMs for Z
    if 0 in povms and isinstance(povms[0], (list, tuple)):
        default_z = povms[0]
    else:
        default_z = None

    # Format remaining POVM values and update attribute
    return {key: _generate_povm(val, default_z) for key, val in povms.items()}


def _format_qubit_povms(
    qubit_povms: any,
) -> Dict[Tuple[int, ...], Dict[Tuple[int, ...], List[DensityMatrix]]]:
    """Format qubit POVMs dict"""
    povms = {}
    if not qubit_povms:
        return povms

    # Format POVM keys to be a tuple of (qubits, basis)
    for qubits, povm in qubit_povms.items():
        if isinstance(qubits, int):
            qubits = (qubits,)

        # Convert value to dict if not already
        if not isinstance(povm, dict):
            formatted_povm = {(i,): val for i, val in enumerate(povm)}
        else:
            # Format dict keys
            formatted_povm = {}
            for index, value in povm.items():
                if isinstance(index, int):
                    index = (index,)
                formatted_povm[tuple(index)] = value

        # Add qubit POVM dict to povms
        povms[qubits] = formatted_povm

    # Format POVM values
    for qubits, povm in povms.items():

        # Convert any Z-povm value if present
        key0 = len(qubits) * (0,)
        if key0 in povm and isinstance(povm[key0], (list, tuple)):
            default_z = povm[key0]
        else:
            default_z = None

        # Convert any values from instructions/channels
        # By applying to Z-povm
        for key, value in povm.items():
            povm[key] = _generate_povm(value, default_z)

    return povms


def _generate_state(
    value: Union[Statevector, DensityMatrix, Instruction, Operator, SuperOp],
    init_state: Optional[QuantumState] = None,
    dims: Optional[Tuple[int, ...]] = None,
) -> DensityMatrix:
    """Format a state into list of DensityMatrix"""
    # If already a quantum state convert to a density matrix
    if isinstance(value, (QuantumState, np.ndarray, list)):
        return DensityMatrix(value, dims=dims)

    # Otherwise convert from operator/channel to POVM effects
    try:
        chan = Operator(value)
    except QiskitError:
        chan = SuperOp(value)
    if dims is None:
        dims = chan.input_dims()

    # Default |0> state density matrix
    if init_state is None:
        init_state = DensityMatrix.from_int(0, dims)
    elif not isinstance(init_state, DensityMatrix):
        init_state = DensityMatrix(init_state, dims)
    return init_state.evolve(chan)


def _format_states(
    states: Optional[Union[List[any], Dict[Tuple[int, ...], any]]],
    init_key: Tuple[int, ...] = (0,),
    instructions: Optional[Sequence[Instruction]] = None,
) -> Dict[Tuple[int, ...], DensityMatrix]:
    "Format default state data"
    # Parse data into dict to include legacy handling
    states = _format_data_dict(states, instructions)

    # Look for default state
    init_val = states.get(init_key, None)
    if isinstance(init_val, (QuantumState, np.ndarray, list)):
        init_state = DensityMatrix(init_val)
    else:
        init_state = None

    # Format remaining states and update attribute
    return {key: _generate_state(val, init_state) for key, val in states.items()}


def _format_qubit_states(
    qubit_states: any,
) -> Dict[Tuple[int, ...], Dict[Tuple[int, ...], DensityMatrix]]:
    """Format qubit POVMs dict"""
    if not qubit_states:
        return {}

    # Format POVM keys to be a tuple of (qubits, basis)
    formatted_states = {}
    for qubits, states in qubit_states.items():
        if isinstance(qubits, int):
            qubits = (qubits,)
        else:
            qubits = tuple(qubits)
        init_key = len(qubits) * (0,)
        formatted_states[qubits] = _format_states(states, init_key)

    return formatted_states


def _format_data_dict(
    data: Optional[any], instructions: Optional[Sequence[Union[Instruction, BaseOperator]]] = None
) -> Dict[Tuple[int, ...], any]:
    "Format default state data"
    # Parse data into dict to include legacy handling
    if isinstance(data, (list, tuple, np.ndarray)):
        iter_data = enumerate(data)
    elif isinstance(data, dict):
        iter_data = data.items()
    elif not data:
        iter_data = {}.items()
    else:
        iter_data = data

    # Format arg to tuple keys
    dict_data = {((key,) if isinstance(key, int) else tuple(key)): val for key, val in iter_data}

    # Add instructions for unspecified data
    if instructions and len(dict_data) < len(instructions):
        for i, inst in enumerate(instructions):
            if (i,) not in dict_data:
                dict_data[(i,)] = inst

    return dict_data



# This code is part of Qiskit.
#
# (C) Copyright IBM 2021.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

"""Device component classes."""

from abc import ABC, abstractmethod


class DeviceComponent(ABC):
    """Abstract class representing a device component. Custom components should be subclassed from
    this class."""

    @abstractmethod
    def __str__(self):
        pass

    def __repr__(self):
        return f"<{self.__class__.__name__}({str(self)})>"

    def __eq__(self, value):
        return str(self) == str(value)


class Qubit(DeviceComponent):
    """Class representing a qubit device component."""

    def __init__(self, index: int):
        self.index = index

    def __str__(self):
        return f"Q{self.index}"

    def __json_encode__(self):
        return {"index": self.index}


class Resonator(DeviceComponent):
    """Class representing a resonator device component."""

    def __init__(self, index: int):
        self.index = index

    def __str__(self):
        return f"R{self.index}"

    def __json_encode__(self):
        return {"index": self.index}


class UnknownComponent(DeviceComponent):
    """Class representing an unknown device component."""

    def __init__(self, component: str):
        self.component = component

    def __str__(self):
        return self.component

    def __json_encode__(self):
        return {"component": self.component}


def to_component(string: str) -> DeviceComponent:
    """Convert the input string to a :class:`.DeviceComponent` instance.

    Args:
        string: String to be converted.

    Returns:
        A :class:`.DeviceComponent` instance.

    Raises:
        ValueError: If input string is not a valid device component.
    """
    if isinstance(string, DeviceComponent):
        return string
    if string.startswith("Q"):
        return Qubit(int(string[1:]))
    if string.startswith("R"):
        return Resonator(int(string[1:]))
    return UnknownComponent(string)



# This code is part of Qiskit.
#
# (C) Copyright IBM 2022.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.
"""
T2Hahn Echo Experiment class.
"""

from typing import List, Optional, Union, Sequence
import numpy as np

from qiskit import QuantumCircuit, QiskitError
from qiskit.circuit import Parameter
from qiskit.providers.backend import Backend

from qiskit_experiments.framework import BackendTiming, BaseExperiment, Options
from qiskit_experiments.library.characterization.analysis.t2hahn_analysis import T2HahnAnalysis


class T2Hahn(BaseExperiment):
    r"""An experiment to measure the dephasing time insensitive to inhomogeneous
    broadening using Hahn echos.

    # section: overview

        This experiment is used to estimate the :math:`T_2` time of a single qubit.
        :math:`T_2` is the dephasing time or the transverse relaxation time of the qubit
        on the Bloch sphere as a result of both energy relaxation and pure dephasing in
        the transverse plane. Unlike :math:`T_2^*`, which is measured by
        :class:`.T2Ramsey`, :math:`T_2` is insensitive to inhomogenous broadening.

        This experiment consists of a series of circuits of the form


        .. parsed-literal::

                 ┌─────────┐┌──────────┐┌───────┐┌──────────┐┌─────────┐┌─┐
            q_0: ┤ Rx(π/2) ├┤ DELAY(t) ├┤ RX(π) ├┤ DELAY(t) ├┤ RX(π/2) ├┤M├
                 └─────────┘└──────────┘└───────┘└──────────┘└─────────┘└╥┘
            c: 1/════════════════════════════════════════════════════════╩═
                                                                         0

        for each *t* from the specified delay times
        and the delays are specified by the user.
        The delays that are specified are delay for each delay gate while
        the delay in the metadata is the total delay which is delay * (num_echoes +1)
        The circuits are run on the device or on a simulator backend.

    # section: manual
        :doc:`/manuals/characterization/t2hahn`

    # section: analysis_ref
        :class:`T2HahnAnalysis`

    # section: example
        .. jupyter-execute::
            :hide-code:

            # backend
            from qiskit_experiments.test.t2hahn_backend import T2HahnBackend

            conversion_factor = 1e-6
            estimated_t2hahn = 20*conversion_factor
            backend = T2HahnBackend(
                t2hahn=[estimated_t2hahn],
                frequency=[100100],
                readout0to1 = [0.02],
                readout1to0 = [0.02],
                )

        .. jupyter-execute::

            import numpy as np
            from qiskit_experiments.library.characterization.t2hahn import T2Hahn

            delays = np.linspace(0, 50, 51)*1e-6

            exp = T2Hahn(physical_qubits=(0, ),
                         delays=delays,
                         backend=backend)
            exp.analysis.set_options(p0=None, plot=True)

            exp_data = exp.run().block_for_results()
            display(exp_data.figure(0))
            exp_data.analysis_results(dataframe=True)

    # section: reference
        .. ref_arxiv:: 1 1904.06560
    """

    @classmethod
    def _default_experiment_options(cls) -> Options:
        """Default experiment options.

        Experiment Options:
            delays (Iterable[float]): Delay times of the experiments.
            num_echoes (int): The number of echoes to preform.
        """
        options = super()._default_experiment_options()

        options.delays = None
        options.num_echoes = 1
        return options

    def __init__(
        self,
        physical_qubits: Sequence[int],
        delays: Union[List[float], np.array],
        num_echoes: int = 1,
        backend: Optional[Backend] = None,
    ):
        """
        Initialize the T2 - Hahn Echo class.

        Args:
            physical_qubits: a single-element sequence containing the qubit whose T2 is to be
                estimated.
            delays: Total delay times of the experiments.
            backend: Optional, the backend to run the experiment on.
            num_echoes: The number of echoes to preform.
            backend: Optional, the backend to run the experiment on.

        Raises:
            QiskitError : Error for invalid input.
        """
        # Initialize base experiment
        super().__init__(physical_qubits, analysis=T2HahnAnalysis(), backend=backend)

        # Set experiment options
        self.set_experiment_options(delays=delays, num_echoes=num_echoes)
        self._verify_parameters()

    def _verify_parameters(self):
        """
        Verify input correctness, raise QiskitError if needed.

        Raises:
            QiskitError : Error for invalid input.
        """
        if any(delay < 0 for delay in self.experiment_options.delays):
            raise QiskitError(
                f"The lengths list {self.experiment_options.delays} should only contain "
                "non-negative elements."
            )

    def circuits(self) -> List[QuantumCircuit]:
        """
        Return a list of experiment circuits.

        Each circuit consists of RX(π/2) followed by a sequence of delay gate,
        RX(π) for echo and delay gate again.
        The sequence repeats for the number of echoes and terminates with RX(±π/2).

        Returns:
            The experiment circuits.
        """
        timing = BackendTiming(self.backend)

        delay_param = Parameter("delay")

        num_echoes = self.experiment_options.num_echoes

        # First X rotation in 90 degrees
        template = QuantumCircuit(1, 1)
        template.rx(np.pi / 2, 0)  # Brings the qubit to the X Axis
        if num_echoes == 0:
            # if number of echoes is 0 then just apply the delay gate
            template.delay(delay_param, 0, timing.delay_unit)
        else:
            for _ in range(num_echoes):
                template.delay(delay_param, 0, timing.delay_unit)
                template.rx(np.pi, 0)
                template.delay(delay_param, 0, timing.delay_unit)

        if num_echoes % 2 == 1:
            template.rx(np.pi / 2, 0)  # X90 again since the num of echoes is odd
        else:
            template.rx(-np.pi / 2, 0)  # X(-90) again since the num of echoes is even
        template.measure(0, 0)  # measure

        circuits = []
        for delay in self.experiment_options.delays:
            if num_echoes == 0:
                single_delay = timing.delay_time(time=delay)
                total_delay = single_delay
            else:
                # Equal delay is put before and after each echo, so each echo gets
                # two delay gates. When there are multiple echoes, the total delay
                # between echoes is 2 * single_delay, made up of two delay gates.
                single_delay = timing.delay_time(time=delay / num_echoes / 2)
                total_delay = single_delay * num_echoes * 2

            assigned = template.assign_parameters(
                {delay_param: timing.round_delay(time=single_delay)}, inplace=False
            )
            assigned.metadata = {"xval": total_delay}
            circuits.append(assigned)

        return circuits

    def _metadata(self):
        metadata = super()._metadata()
        # Store measurement level and meas return if they have been
        # set for the experiment
        for run_opt in ["meas_level", "meas_return"]:
            if hasattr(self.run_options, run_opt):
                metadata[run_opt] = getattr(self.run_options, run_opt)
        return metadata



# This code is part of Qiskit.
#
# (C) Copyright IBM 2021.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.
"""
Quantum process tomography analysis
"""


from typing import List, Union, Callable
from collections import defaultdict
import numpy as np
import scipy.linalg as la
from uncertainties import ufloat

from qiskit.quantum_info import DensityMatrix, Choi, Operator
from qiskit.quantum_info.operators.base_operator import BaseOperator
from qiskit.quantum_info.operators.channel.quantum_channel import QuantumChannel

from qiskit_experiments.exceptions import AnalysisError
from qiskit_experiments.framework import BaseAnalysis, AnalysisResultData, Options
from qiskit_experiments.framework.package_deps import version_is_at_least
from .fitters import (
    tomography_fitter_data,
    postprocess_fitter,
    linear_inversion,
    cvxpy_linear_lstsq,
    cvxpy_gaussian_lstsq,
)


class TomographyAnalysis(BaseAnalysis):
    """Base analysis for state and process tomography experiments."""

    _builtin_fitters = {
        "linear_inversion": linear_inversion,
        "cvxpy_linear_lstsq": cvxpy_linear_lstsq,
        "cvxpy_gaussian_lstsq": cvxpy_gaussian_lstsq,
    }
    _cvxpy_fitters = (
        cvxpy_linear_lstsq,
        cvxpy_gaussian_lstsq,
    )

    @classmethod
    def _default_options(cls) -> Options:
        """Default analysis options

        Analysis Options:
            measurement_basis
                (:class:`~qiskit_experiments.library.tomography.basis.MeasurementBasis`):
                The measurement
                :class:`~qiskit_experiments.library.tomography.basis.MeasurementBasis`
                to use for tomographic reconstruction when running a
                :class:`~qiskit_experiments.library.tomography.StateTomography` or
                :class:`~qiskit_experiments.library.tomography.ProcessTomography`.
            preparation_basis
                (:class:`~qiskit_experiments.library.tomography.basis.PreparationBasis`):
                The preparation
                :class:`~qiskit_experiments.library.tomography.basis.PreparationBasis`
                to use for tomographic reconstruction for
                :class:`~qiskit_experiments.library.tomography.ProcessTomography`.
            fitter (str or Callable): The fitter function to use for reconstruction.
                This can  be a string to select one of the built-in fitters, or a callable to
                supply a custom fitter function. See the `Fitter Functions` section for
                additional information.
            fitter_options (dict): Any addition kwarg options to be supplied to the fitter
                function. For documentation of available kwargs refer to the fitter function
                documentation.
            rescale_positive (bool): If True rescale the state returned by the fitter
                to be positive-semidefinite. See the `PSD Rescaling` section for
                additional information (Default: True).
            rescale_trace (bool): If True rescale the state returned by the fitter
                have either trace 1 for :class:`~qiskit.quantum_info.DensityMatrix`,
                or trace dim for :class:`~qiskit.quantum_info.Choi` matrices (Default: True).
            measurement_qubits (Sequence[int]): Optional, the physical qubits with tomographic
                measurements. If not specified will be set to ``[0, ..., N-1]`` for N-qubit
                tomographic measurements.
            preparation_qubits (Sequence[int]): Optional, the physical qubits with tomographic
                preparations. If not specified will be set to ``[0, ..., N-1]`` for N-qubit
                tomographic preparations.
            target (Any): Optional, target object for fidelity comparison of the fit
                (Default: None).
            target_bootstrap_samples (int): Optional, number of outcome re-samples to draw
                from measurement data for each basis for computing a bootstrapped standard
                error of fidelity with the target state. If 0 no bootstrapping will be
                performed and the target fidelity will not include a standard error
                (Default: 0).
            target_bootstrap_seed (int | None | Generator): Optional, RNG seed or
                Generator to use for bootstrapping data for boostrapped fidelity
                standard error calculation (Default: None).
            conditional_circuit_clbits (list[int]): Optional, the clbit indices in the
                source circuit to be conditioned on when reconstructing the state.
                Enabling this will return a list of reconstructed state components
                conditional on the values of these clbit values. The integer value of the
                conditioning clbits is stored in state analysis result extra field
                `"conditional_circuit_outcome"`.
            conditional_measurement_indices (list[int]): Optional, indices of tomography
                measurement qubits to used for conditional state reconstruction. Enabling
                this will return a list of reconstructed state components conditioned on
                the remaining tomographic bases conditional on the basis index, and outcome
                value for these measurements. The conditional measurement basis index and
                integer value of the measurement outcome is stored in state analysis result
                extra fields `"conditional_measurement_index"` and
                `"conditional_measurement_outcome"` respectively.
            conditional_preparation_indices (list[int]): Optional, indices of tomography
                preparation qubits to used for conditional state reconstruction. Enabling
                this will return a list of reconstructed channel components conditioned on
                the remaining tomographic bases conditional on the basis index. The
                conditional preparation basis index is stored in state analysis result
                extra fields `"conditional_preparation_index"`.
            extra (Dict[str, Any]): Extra metadata dictionary attached to analysis results.
        """
        options = super()._default_options()

        options.measurement_basis = None
        options.preparation_basis = None
        options.fitter = "linear_inversion"
        options.fitter_options = {}
        options.rescale_positive = True
        options.rescale_trace = True
        options.measurement_qubits = None
        options.preparation_qubits = None
        options.target = None
        options.target_bootstrap_samples = 0
        options.target_bootstrap_seed = None
        options.conditional_circuit_clbits = None
        options.conditional_measurement_indices = None
        options.conditional_preparation_indices = None
        options.extra = {}
        return options

    @classmethod
    def _get_fitter(cls, fitter: Union[str, Callable]) -> Callable:
        """Return fitter function for named builtin fitters"""
        if fitter is None:
            raise AnalysisError("No tomography fitter given")
        if not isinstance(fitter, str):
            return fitter
        if fitter in cls._builtin_fitters:
            return cls._builtin_fitters[fitter]
        raise AnalysisError(f"Unrecognized tomography fitter {fitter}")

    def _run_analysis(self, experiment_data):

        # Get option values
        meas_basis = self.options.measurement_basis
        meas_qubits = self.options.measurement_qubits
        if meas_basis and meas_qubits is None:
            meas_qubits = experiment_data.metadata.get("m_qubits")
        prep_basis = self.options.preparation_basis
        prep_qubits = self.options.preparation_qubits
        if prep_basis and prep_qubits is None:
            prep_qubits = experiment_data.metadata.get("p_qubits")
        cond_meas_indices = self.options.conditional_measurement_indices
        if cond_meas_indices is True:
            cond_meas_indices = list(range(len(meas_qubits)))
        cond_prep_indices = self.options.conditional_preparation_indices
        if cond_prep_indices is True:
            cond_prep_indices = list(range(len(prep_qubits)))

        # Generate tomography fitter data
        outcome_shape = None
        if meas_basis and meas_qubits:
            outcome_shape = meas_basis.outcome_shape(meas_qubits)

        outcome_data, shot_data, meas_data, prep_data = tomography_fitter_data(
            experiment_data.data(),
            outcome_shape=outcome_shape,
        )
        qpt = prep_data.size > 0

        # Get fitter kwargs
        fitter_kwargs = {}
        if meas_basis:
            fitter_kwargs["measurement_basis"] = meas_basis
        if meas_qubits:
            fitter_kwargs["measurement_qubits"] = meas_qubits
        if cond_meas_indices:
            fitter_kwargs["conditional_measurement_indices"] = cond_meas_indices
        if prep_basis:
            fitter_kwargs["preparation_basis"] = prep_basis
        if prep_qubits:
            fitter_kwargs["preparation_qubits"] = prep_qubits
        if cond_prep_indices:
            fitter_kwargs["conditional_preparation_indices"] = cond_prep_indices
        fitter_kwargs.update(**self.options.fitter_options)
        fitter = self._get_fitter(self.options.fitter)

        # Fit state results
        state_results = self._fit_state_results(
            fitter,
            outcome_data,
            shot_data,
            meas_data,
            prep_data,
            qpt=qpt,
            **fitter_kwargs,
        )

        other_results = []

        # Compute fidelity with target
        if len(state_results) == 1:
            other_results += self._fidelity_result(
                state_results[0],
                fitter,
                outcome_data,
                shot_data,
                meas_data,
                prep_data,
                qpt=qpt,
                **fitter_kwargs,
            )

        # Check positive
        other_results += self._positivity_result(state_results, qpt=qpt)

        # Check trace preserving
        if qpt:
            output_dim = np.prod(state_results[0].value.output_dims())
            other_results += self._tp_result(state_results, output_dim)

        # Finally format state result metadata to remove eigenvectors
        # which are no longer needed to reduce size
        for state_result in state_results:
            state_result.extra.pop("eigvecs")

        analysis_results = state_results + other_results

        if self.options.extra:
            for res in analysis_results:
                res.extra.update(self.options.extra)
        return analysis_results, []

    def _fit_state_results(
        self,
        fitter: Callable,
        outcome_data: np.ndarray,
        shot_data: np.ndarray,
        measurement_data: np.ndarray,
        preparation_data: np.ndarray,
        qpt: Union[bool, str, None] = "auto",
        **fitter_kwargs,
    ):
        """Fit state results from tomography data,"""
        try:
            fits, fitter_metadata = fitter(
                outcome_data,
                shot_data,
                measurement_data,
                preparation_data,
                **fitter_kwargs,
            )
        except AnalysisError as ex:
            raise AnalysisError(f"Tomography fitter failed with error: {str(ex)}") from ex

        # Post process fit
        states, states_metadata = postprocess_fitter(
            fits,
            fitter_metadata,
            make_positive=self.options.rescale_positive,
            trace="auto" if self.options.rescale_trace else None,
            qpt=qpt,
        )

        # Convert to results
        state_results = [
            AnalysisResultData("state", state, extra=extra)
            for state, extra in zip(states, states_metadata)
        ]
        return state_results

    def _fidelity_result(
        self,
        state_result: AnalysisResultData,
        fitter: Callable,
        outcome_data: np.ndarray,
        shot_data: np.ndarray,
        measurement_data: np.ndarray,
        preparation_data: np.ndarray,
        qpt: bool = False,
        **fitter_kwargs,
    ) -> List[AnalysisResultData]:
        """Calculate fidelity result if a target has been set"""
        target = self.options.target
        if target is None:
            return []

        # Compute fidelity
        name = "process_fidelity" if qpt else "state_fidelity"
        fidelity = self._compute_fidelity(state_result, target, qpt=qpt)

        if not self.options.target_bootstrap_samples:
            # No bootstrapping
            return [AnalysisResultData(name, fidelity)]

        # Optionally, Estimate std error of fidelity via boostrapping
        seed = self.options.target_bootstrap_seed
        if isinstance(seed, np.random.Generator):
            rng = seed
        else:
            rng = np.random.default_rng(seed)
        prob_data = outcome_data / shot_data[None, :, None]
        bs_fidelities = []
        for _ in range(self.options.target_bootstrap_samples):
            # TODO: remove conditional once numpy is pinned at 1.22 and above
            if version_is_at_least("numpy", "1.22"):
                sampled_data = rng.multinomial(shot_data, prob_data)
            else:
                sampled_data = np.zeros_like(outcome_data)
                for i in range(prob_data.shape[0]):
                    for j in range(prob_data.shape[1]):
                        sampled_data[i, j] = rng.multinomial(shot_data[j], prob_data[i, j])

            try:
                state_results = self._fit_state_results(
                    fitter,
                    sampled_data,
                    shot_data,
                    measurement_data,
                    preparation_data,
                    qpt=qpt,
                    **fitter_kwargs,
                )
                bs_fidelities.append(self._compute_fidelity(state_results[0], target, qpt=qpt))
            except AnalysisError:
                pass

        bs_stderr = np.std(bs_fidelities)
        return [
            AnalysisResultData(
                name,
                ufloat(fidelity, bs_stderr),
                extra={"bootstrap_samples": bs_fidelities},
            )
        ]

    @staticmethod
    def _positivity_result(
        state_results: List[AnalysisResultData], qpt: bool = False
    ) -> List[AnalysisResultData]:
        """Check if eigenvalues are positive"""
        total_cond = defaultdict(float)
        comps_cond = defaultdict(list)
        comps_pos = defaultdict(list)
        name = "completely_positive" if qpt else "positive"
        for result in state_results:
            cond_idx = result.extra.get("conditional_measurement_index", None)
            evals = result.extra["eigvals"]

            # Check if component is positive and add to extra if so
            cond = np.sum(np.abs(evals[evals < 0]))
            pos = bool(np.isclose(cond, 0))
            result.extra[name] = pos

            # Add component to combined result
            comps_cond[cond_idx].append(cond)
            comps_pos[cond_idx].append(pos)
            total_cond[cond_idx] += cond * result.extra["conditional_probability"]

        # Check if combined conditional state is positive
        results = []
        for key, delta in total_cond.items():
            is_pos = bool(np.isclose(delta, 0))
            result = AnalysisResultData(name, is_pos)
            if not is_pos:
                result.extra = {
                    "delta": delta,
                    "components": comps_pos[key],
                    "components_delta": comps_cond[key],
                }
            if key:
                result.extra["conditional_measurement_index"] = key
            results.append(result)
        return results

    @staticmethod
    def _tp_result(
        state_results: List[AnalysisResultData],
        output_dim: int = 1,
    ) -> List[AnalysisResultData]:
        """Check if QPT channel is trace preserving"""
        # Construct the Kraus TP condition matrix sum_i K_i^dag K_i
        # summed over all components k
        kraus_cond = {}
        for result in state_results:
            evals = result.extra["eigvals"]
            evecs = result.extra["eigvecs"]
            prob = result.extra["conditional_probability"]
            cond_meas_idx = result.extra.get("conditional_measurement_index", None)
            cond_prep_idx = result.extra.get("conditional_preparation_index", None)
            cond_idx = (cond_prep_idx, cond_meas_idx)
            size = len(evals)
            input_dim = size // output_dim
            mats = np.reshape(evecs.T, (size, input_dim, output_dim), order="F")
            comp_cond = np.einsum("i,iaj,ibj->ab", evals, mats.conj(), mats)
            if cond_idx in kraus_cond:
                kraus_cond[cond_idx] += prob * comp_cond
            else:
                kraus_cond[cond_idx] = prob * comp_cond

        results = []
        for key, val in kraus_cond.items():
            tp_cond = np.sum(np.abs(la.eigvalsh(val - np.eye(input_dim))))
            is_tp = bool(np.isclose(tp_cond, 0, atol=1e-5))
            result = AnalysisResultData("trace_preserving", is_tp, extra={})
            if not is_tp:
                result.extra["delta"] = tp_cond
            if key:
                result.extra["conditional_measurement_index"] = key
            results.append(result)
        return results

    @staticmethod
    def _compute_fidelity(
        state_result: AnalysisResultData,
        target: Union[Choi, DensityMatrix],
        qpt: bool = False,
    ) -> AnalysisResultData:
        """Faster computation of fidelity from eigen decomposition"""
        if qpt:
            input_dim = np.prod(state_result.value.input_dims())
        else:
            input_dim = 1
        evals = state_result.extra["eigvals"]
        evecs = state_result.extra["eigvecs"]

        # Format target to statevector or densitymatrix array
        if target is None:
            raise AnalysisError("No target state provided")
        if isinstance(target, QuantumChannel):
            target_state = Choi(target).data / input_dim
        elif isinstance(target, BaseOperator):
            target_state = np.ravel(Operator(target), order="F") / np.sqrt(input_dim)
        else:
            # Statevector or density matrix
            target_state = np.array(target)

        if target_state.ndim == 1:
            rho = evecs @ (evals / input_dim * evecs).T.conj()
            fidelity = np.real(target_state.conj() @ rho @ target_state)
        else:
            sqrt_rho = evecs @ (np.sqrt(evals / input_dim) * evecs).T.conj()
            eig = la.eigvalsh(sqrt_rho @ target_state @ sqrt_rho)
            fidelity = np.sum(np.sqrt(np.maximum(eig, 0))) ** 2
        return fidelity



# This code is part of Qiskit.
#
# (C) Copyright IBM 2021.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.
"""
Utility functions for CVXPy module
"""

from typing import Callable, List, Tuple, Optional, Union
import functools
import numpy as np
import scipy.sparse as sps

from qiskit_experiments.exceptions import AnalysisError

# Check if CVXPY package is installed
try:
    import cvxpy
    from cvxpy import Problem, Variable
    from cvxpy.constraints.constraint import Constraint

    HAS_CVXPY = True

except ImportError:
    cvxpy = None

    HAS_CVXPY = False

    # Used for type hints
    Problem = None
    Variable = None
    Constraint = None


def requires_cvxpy(func: Callable) -> Callable:
    """Function decorator for functions requiring CVXPy.

    Args:
        func: a function requiring CVXPy.

    Returns:
        The decorated function.

    Raises:
        QiskitError: If CVXPy is not installed.
    """

    @functools.wraps(func)
    def decorated_func(*args, **kwargs):
        if not HAS_CVXPY:
            raise ImportError(
                f"The CVXPY package is required to for {func}."
                "You can install it with 'pip install cvxpy'."
            )
        return func(*args, **kwargs)

    return decorated_func


def solve_iteratively(
    problem: Problem, initial_iters: int, scale: int = 2, solver: str = "SCS", **solve_kwargs
) -> None:
    """Solve a CVXPY problem increasing iterations if solution is inaccurate.

    If the problem is not solved with the ``initial_iters`` value of
    iterations the number of iterations will be doubled up to the
    specified ``"max_iters"`` value in the solve_kwargs. If no max
    iters is specified this will be set to 4 times the initial iters
    values.

    Args:
        problem: A CVXPY Problem to solve
        initial_iters: The initial number of max iterations to use
                       when solving the problem
        scale: Scale factor for increasing the initial_iters up to
               max_iters at each step (Default: 2).
        solver: The solver to use. Defaults to the Splitting Conic Solver.
        solve_kwargs: kwargs for problem.solve method.

    Raises:
        AnalysisError: If the CVXPY solve fails to return an optimal or
                       optimal_inaccurate solution.
    """
    current_max_iters = initial_iters
    final_max_iters = solve_kwargs.get("max_iters", 2 * scale * initial_iters)
    problem_solved = False
    while not problem_solved:
        solve_kwargs["max_iters"] = current_max_iters
        problem.solve(solver=solver, **solve_kwargs)
        if problem.status in ["optimal_inaccurate", "optimal"]:
            problem_solved = True
        elif problem.status == "unbounded_inaccurate":
            if scale > 1 and current_max_iters < final_max_iters:
                current_max_iters = int(scale * current_max_iters)
            else:
                raise AnalysisError(
                    "CVXPY solver failed to find an optimal solution in "
                    "the given number of iterations. Try setting a larger "
                    "value for 'max_iters' solver option."
                )
        else:
            raise AnalysisError(f"CVXPY solver failed with problem status '{problem.status}'.")


def complex_matrix_variable(
    dim: int, hermitian: bool = False, psd: bool = False, trace: Optional[complex] = None
) -> Tuple[Variable, Variable, List[Constraint]]:
    """Construct a pair of real variables and constraints for a Hermitian matrix

    Args:
        dim: The dimension of the complex square matrix.
        hermitian: If True add constraint that the matrix is Hermitian.
                   (Default: False).
        psd: If True add a constraint that the matrix is positive
             semidefinite (Default: False).
        trace: Optional, add a constraint that the trace of the matrix is
               the specified value.

    Returns:
        A tuple ``(mat.real, mat.imag, constraints)`` of two real CVXPY
        matrix variables, and constraints.
    """
    mat_r = cvxpy.Variable((dim, dim))
    mat_i = cvxpy.Variable((dim, dim))
    cons = []

    if hermitian:
        cons += hermitian_constraint(mat_r, mat_i)
    if trace is not None:
        cons += trace_constraint(mat_r, mat_i, trace)
    if psd:
        cons += psd_constraint(mat_r, mat_i)
    return mat_r, mat_i, cons


def hermitian_constraint(mat_r: Variable, mat_i: Variable) -> List[Constraint]:
    """Return CVXPY constraint for a Hermitian matrix variable.

    Args:
        mat_r: The CVXPY variable for the real part of the matrix.
        mat_i: The CVXPY variable for the complex part of the matrix.

    Returns:
        A list of constraints on the real and imaginary parts.
    """
    return [mat_r == mat_r.T, mat_i == -mat_i.T]


def psd_constraint(mat_r: Variable, mat_i: Variable) -> List[Constraint]:
    """Return CVXPY Hermitian constraints for a complex matrix.

    Args:
        mat_r: The CVXPY variable for the real part of the matrix.
        mat_i: The CVXPY variable for the complex part of the matrix.

    Returns:
        A list of constraints on the real and imaginary parts.
    """
    bmat = cvxpy.bmat([[mat_r, -mat_i], [mat_i, mat_r]])
    return [bmat >> 0]


def trace_constraint(
    mat_r: Union[Variable, List[Variable]],
    mat_i: Union[Variable, List[Variable]],
    trace: complex,
    hermitian: bool = False,
) -> List[Constraint]:
    """Return CVXPY trace constraints for a complex matrix.

    Args:
        mat_r: The CVXPY variable for the real part of the matrix.
        mat_i: The CVXPY variable for the complex part of the matrix.
        trace: The value for the trace constraint.
        hermitian: If the input variables are Hermitian, only the real trace constraint
                   is required.
    Returns:
        A list of constraints on the real and imaginary parts.

    Raises:
        TypeError: If input variables are not valid.
    """
    if isinstance(mat_r, (list, tuple)):
        arg_r = cvxpy.sum(mat_r)
    elif isinstance(mat_r, Variable):
        arg_r = mat_r
    else:
        raise TypeError("Input must be a cvxpy variable or list of variables")
    cons = [cvxpy.trace(arg_r) == np.real(trace)]

    if hermitian:
        return cons

    # If not hermitian add imaginary trace constraint
    if isinstance(mat_i, (list, tuple)):
        arg_i = cvxpy.sum(mat_i)
    elif isinstance(mat_i, Variable):
        arg_i = mat_i
    else:
        raise TypeError("Input must be a cvxpy variable or list of variables")
    cons.append(cvxpy.trace(arg_i) == np.imag(trace))

    return cons


def partial_trace_constaint(
    mat_r: Variable,
    mat_i: Variable,
    constraint: np.ndarray,
) -> List[Constraint]:
    """Return CVXPY partial trace constraints for a complex matrix.

    Args:
        mat_r: The CVXPY variable for the real part of the matrix.
        mat_i: The CVXPY variable for the complex part of the matrix.
        constraint: The constraint matrix to set the partial trace to.

    Returns:
        A list of constraints on the real and imaginary parts.

    Raises:
        TypeError: If input variables are not valid.
    """
    sdim = mat_r.shape[0]
    output_dim = constraint.shape[0]
    input_dim = sdim // output_dim
    ptr = partial_trace_super(input_dim, output_dim)
    vec_cons = np.ravel(constraint, order="F")
    return [
        ptr @ cvxpy.vec(mat_r) == vec_cons.real.round(12),
        ptr @ cvxpy.vec(mat_i) == vec_cons.imag.round(12),
    ]


def trace_preserving_constaint(
    mat_r: Union[Variable, List[Variable]],
    mat_i: Union[Variable, List[Variable]],
    input_dim: Optional[int] = None,
    hermitian: bool = False,
) -> List[Constraint]:
    """Return CVXPY trace preserving constraints for a complex matrix.

    Args:
        mat_r: The CVXPY variable for the real part of the matrix.
        mat_i: The CVXPY variable for the complex part of the matrix.
        input_dim: Optional, the input dimension for the system channel if the input
                   and output dimensions are not equal.
        hermitian: If the input variables are Hermitian, only the real trace constraint
                   is required.

    Returns:
        A list of constraints on the real and imaginary parts.

    Raises:
        TypeError: If input variables are not valid.
    """
    if isinstance(mat_r, (tuple, list)):
        sdim = mat_r[0].shape[0]
        arg_r = cvxpy.sum(mat_r)
    elif isinstance(mat_r, Variable):
        sdim = mat_r.shape[0]
        arg_r = mat_r
    else:
        raise TypeError("Input must be a cvxpy variable or list of variables")
    if input_dim is None:
        input_dim = int(np.sqrt(sdim))
    output_dim = sdim // input_dim

    ptr = partial_trace_super(input_dim, output_dim)
    cons = [ptr @ cvxpy.vec(arg_r) == np.identity(input_dim).ravel()]

    if hermitian:
        return cons

    # If not hermitian add imaginary partial trace constraint
    if isinstance(mat_i, (tuple, list)):
        arg_i = cvxpy.sum(mat_i)
    elif isinstance(mat_i, Variable):
        arg_i = mat_i
    else:
        raise TypeError("Input must be a cvxpy variable or list of variables")
    cons.append(ptr @ cvxpy.vec(arg_i) == np.zeros(input_dim**2))
    return cons


@functools.lru_cache(3)
def partial_trace_super(dim1: int, dim2: int) -> np.array:
    """
    Return the partial trace superoperator in the column-major basis.

    This returns the superoperator S_TrB such that:
        S_TrB * vec(rho_AB) = vec(rho_A)
    for rho_AB = kron(rho_A, rho_B)

    Args:
        dim1: the dimension of the system not being traced
        dim2: the dimension of the system being traced over

    Returns:
        A Numpy array of the partial trace superoperator S_TrB.
    """
    iden = sps.identity(dim1)
    ptr = sps.csr_matrix((dim1 * dim1, dim1 * dim2 * dim1 * dim2))

    for j in range(dim2):
        v_j = sps.coo_matrix(([1], ([0], [j])), shape=(1, dim2))
        tmp = sps.kron(iden, v_j.tocsr())
        ptr += sps.kron(tmp, tmp)

    return ptr



# This code is part of Qiskit.
#
# (C) Copyright IBM 2023.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

"""Container of experiment data components."""

from __future__ import annotations

import copy
import io
from typing import Dict, Optional, Union, Any

from matplotlib.figure import Figure as MatplotlibFigure


class FigureData:
    """A plot data container.

    .. note::
        Raw figure data can be accessed through the :attr:`.FigureData.figure` attribute.

    """

    def __init__(
        self,
        figure,
        name: str | None = None,
        metadata: dict[str, Any] | None = None,
    ):
        """Creates a new figure data object.

        Args:
            figure: The raw figure itself. Can be SVG or matplotlib.Figure.
            name: The name of the figure.
            metadata: Any metadata to be stored with the figure.
        """
        self.figure = figure
        self._name = name
        self.metadata = metadata or {}

    def __eq__(self, value):
        """Test equality between two instances of FigureData."""
        return vars(self) == vars(value)

    # name is read only
    @property
    def name(self) -> str:
        """The name of the figure"""
        return self._name

    @property
    def metadata(self) -> dict:
        """The metadata dictionary stored with the figure"""
        return self._metadata

    @metadata.setter
    def metadata(self, new_metadata: dict):
        """Set the metadata to new value; must be a dictionary"""
        if not isinstance(new_metadata, dict):
            raise ValueError("figure metadata must be a dictionary")
        self._metadata = new_metadata

    def copy(self, new_name: Optional[str] = None):
        """Creates a copy of the figure data"""
        name = new_name or self.name
        return FigureData(figure=self.figure, name=name, metadata=copy.deepcopy(self.metadata))

    def __json_encode__(self) -> Dict[str, Any]:
        """Return the json representation of the figure data"""
        return {"figure": self.figure, "name": self.name, "metadata": self.metadata}

    @classmethod
    def __json_decode__(cls, args: Dict[str, Any]) -> "FigureData":
        """Initialize a figure data from the json representation"""
        return cls(**args)

    def _repr_png_(self):
        if isinstance(self.figure, MatplotlibFigure):
            b = io.BytesIO()
            self.figure.savefig(b, format="png", bbox_inches="tight")
            png = b.getvalue()
            return png
        else:
            return None

    def _repr_svg_(self):
        if isinstance(self.figure, str):
            return self.figure
        if isinstance(self.figure, bytes):
            return self.figure.decode("utf-8")
        return None


FigureType = Union[str, bytes, MatplotlibFigure, FigureData]



# This code is part of Qiskit.
#
# (C) Copyright IBM 2023.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.
"""P1 experiment at various qubit frequencies."""

from __future__ import annotations

import warnings
from collections.abc import Sequence

import numpy as np
from qiskit import pulse
from qiskit.circuit import QuantumCircuit, Gate, Parameter, ParameterExpression
from qiskit.providers.backend import Backend
from qiskit.utils import optionals as _optional
from qiskit.utils.deprecation import deprecate_func

from qiskit_experiments.framework import BackendTiming, BaseExperiment, Options
from .p1_spect_analysis import StarkP1SpectAnalysis

from .coefficients import (
    StarkCoefficients,
    retrieve_coefficients_from_backend,
)

if _optional.HAS_SYMENGINE:
    import symengine as sym
else:
    import sympy as sym


class StarkP1Spectroscopy(BaseExperiment):
    """P1 spectroscopy experiment with Stark tone.

    # section: overview

        This experiment measures a probability of the excitation state of the qubit
        with a certain delay after excitation.
        A Stark tone is applied during this delay to move the
        qubit frequency to conduct a spectroscopy of qubit relaxation quantity.

        .. parsed-literal::

                 ┌───┐┌──────────────────┐┌─┐
              q: ┤ X ├┤ Stark(stark_amp) ├┤M├
                 └───┘└──────────────────┘└╥┘
            c: 1/══════════════════════════╩═
                                           0

        Since the qubit relaxation rate may depend on the qubit frequency due to the
        coupling to nearby energy levels, this experiment is useful to find out
        lossy operation frequency that might be harmful to the gate fidelity [1].

    # section: analysis_ref
        :class:`qiskit_experiments.library.driven_freq_tuning.StarkP1SpectAnalysis`

    # section: reference
        .. ref_arxiv:: 1 2105.15201

    # section: see_also
        :class:`qiskit_experiments.library.driven_freq_tuning.ramsey.StarkRamseyXY`
        :class:`qiskit_experiments.library.driven_freq_tuning.ramsey_amp_scan.StarkRamseyXYAmpScan`

    # section: manual
        :doc:`/manuals/characterization/stark_experiment`

    """

    @deprecate_func(
        since="0.8",
        package_name="qiskit-experiments",
        additional_msg=(
            "Due to the deprecation of Qiskit Pulse, experiments involving pulse "
            "gate calibrations like this one have been deprecated."
        ),
    )
    def __init__(
        self,
        physical_qubits: Sequence[int],
        backend: Backend | None = None,
        **experiment_options,
    ):
        """
        Initialize new experiment class.

        Args:
            physical_qubits: Sequence with the index of the physical qubit.
            backend: Optional, the backend to run the experiment on.
            experiment_options: Experiment options. See the class documentation or
                ``self._default_experiment_options`` for descriptions.
        """
        self._timing = None

        with warnings.catch_warnings():
            warnings.filterwarnings(
                "ignore",
                message="deprecation of Qiskit Pulse",
                module="qiskit_experiments",
                category=DeprecationWarning,
            )
            analysis = StarkP1SpectAnalysis()

        super().__init__(
            physical_qubits=physical_qubits,
            analysis=analysis,
            backend=backend,
        )
        self.set_experiment_options(**experiment_options)

    @classmethod
    def _default_experiment_options(cls) -> Options:
        """Default experiment options.

        Experiment Options:
            t1_delay (float): The T1 delay time after excitation pulse. The delay must be
                sufficiently greater than the edge duration determined by the stark_sigma.
            stark_channel (PulseChannel): Pulse channel to apply Stark tones.
                If not provided, the same channel with the qubit drive is used.
            stark_freq_offset (float): Offset of Stark tone frequency from the qubit frequency.
                This must be greater than zero not to apply Rabi drive.
            stark_sigma (float): Gaussian sigma of the rising and falling edges
                of the Stark tone, in seconds.
            stark_risefall (float): Ratio of sigma to the duration of
                the rising and falling edges of the Stark tone.
            min_xval (float): Minimum x value.
            max_xval (float): Maximum x value.
            num_xvals (int): Number of x-values to scan.
            xval_type (str): Type of x-value. Either ``amplitude`` or ``frequency``.
                Setting to frequency requires pre-calibration of Stark shift coefficients.
            spacing (str): A policy for the spacing to create an amplitude list from
                ``min_stark_amp`` to ``max_stark_amp``. Either ``linear`` or ``quadratic``
                must be specified.
            xvals (list[float]): The list of x-values that will be scanned in the experiment.
                If not set, then ``num_xvals`` parameters spaced according to
                the ``spacing`` policy between ``min_xval`` and ``max_xval`` are used.
                If ``xvals`` is set, these parameters are ignored.
            stark_coefficients (StarkCoefficients): Calibrated Stark shift coefficients.
                This value is necessary when xval_type is "frequency".
                When this value is None, a search for the "stark_coefficients" in the
                result database is run. This requires having the experiment service
                available in the backend set for the experiment.
        """
        options = super()._default_experiment_options()
        options.update_options(
            t1_delay=20e-6,
            stark_channel=None,
            stark_freq_offset=80e6,
            stark_sigma=15e-9,
            stark_risefall=2,
            min_xval=-1.0,
            max_xval=1.0,
            num_xvals=201,
            xval_type="amplitude",
            spacing="quadratic",
            xvals=None,
            stark_coefficients=None,
        )
        options.set_validator("spacing", ["linear", "quadratic"])
        options.set_validator("xval_type", ["amplitude", "frequency"])
        options.set_validator("stark_freq_offset", (0, np.inf))
        options.set_validator("stark_channel", pulse.channels.PulseChannel)
        options.set_validator("stark_coefficients", StarkCoefficients)
        return options

    def _set_backend(self, backend: Backend):
        super()._set_backend(backend)
        self._timing = BackendTiming(backend)

    def parameters(self) -> np.ndarray:
        """Stark tone amplitudes to use in circuits.

        Returns:
            The list of amplitudes to use for the different circuits based on the
            experiment options.

        Raises:
            ValueError: When invalid xval spacing is specified.
        """
        opt = self.experiment_options  # alias

        if opt.xvals is None:
            if opt.spacing == "linear":
                params = np.linspace(opt.min_xval, opt.max_xval, opt.num_xvals)
            elif opt.spacing == "quadratic":
                min_sqrt = np.sign(opt.min_xval) * np.sqrt(np.abs(opt.min_xval))
                max_sqrt = np.sign(opt.max_xval) * np.sqrt(np.abs(opt.max_xval))
                lin_params = np.linspace(min_sqrt, max_sqrt, opt.num_xvals)
                params = np.sign(lin_params) * lin_params**2
            else:
                raise ValueError(f"Spacing option {opt.spacing} is not valid.")
        else:
            params = np.asarray(opt.xvals, dtype=float)

        if opt.xval_type == "frequency":
            coeffs = opt.stark_coefficients
            if coeffs is None:
                coeffs = retrieve_coefficients_from_backend(
                    backend=self.backend,
                    qubit=self.physical_qubits[0],
                )
            return coeffs.convert_freq_to_amp(freqs=params)
        return params

    def parameterized_circuits(self) -> tuple[QuantumCircuit, ...]:
        """Create circuits with parameters for P1 experiment with Stark shift.

        Returns:
            Quantum template circuit for P1 experiment.
        """
        opt = self.experiment_options  # alias
        param = Parameter("stark_amp")
        sym_param = param._symbol_expr

        # Pulse gates
        stark = Gate("Stark", 1, [param])

        # Note that Stark tone yields negative (positive) frequency shift
        # when the Stark tone frequency is higher (lower) than qubit f01 frequency.
        # This choice gives positive frequency shift with positive Stark amplitude.
        qubit_f01 = self._backend_data.drive_freqs[self.physical_qubits[0]]
        neg_sign_of_amp = ParameterExpression(
            symbol_map={param: sym_param},
            expr=-sym.sign(sym_param),
        )
        abs_of_amp = ParameterExpression(
            symbol_map={param: sym_param},
            expr=sym.Abs(sym_param),
        )
        stark_freq = qubit_f01 + neg_sign_of_amp * opt.stark_freq_offset
        stark_channel = opt.stark_channel or pulse.DriveChannel(self.physical_qubits[0])
        sigma_dt = opt.stark_sigma / self._backend_data.dt
        delay_dt = self._timing.round_pulse(time=opt.t1_delay)

        with pulse.build() as stark_schedule:
            pulse.set_frequency(stark_freq, stark_channel)
            pulse.play(
                pulse.GaussianSquare(
                    duration=delay_dt,
                    amp=abs_of_amp,
                    sigma=sigma_dt,
                    risefall_sigma_ratio=opt.stark_risefall,
                ),
                stark_channel,
            )

        temp_t1 = QuantumCircuit(1, 1)
        temp_t1.x(0)
        temp_t1.append(stark, [0])
        temp_t1.measure(0, 0)
        temp_t1.add_calibration(
            gate=stark,
            qubits=self.physical_qubits,
            schedule=stark_schedule,
        )

        return (temp_t1,)

    def circuits(self) -> list[QuantumCircuit]:
        """Create circuits.

        Returns:
            A list of P1 circuits with a variable Stark tone amplitudes.
        """
        (t1_circ,) = self.parameterized_circuits()
        param = next(iter(t1_circ.parameters))

        circs = []
        for amp in self.parameters():
            t1_assigned = t1_circ.assign_parameters({param: amp}, inplace=False)
            t1_assigned.metadata = {"xval": amp}
            circs.append(t1_assigned)

        return circs

    def _metadata(self) -> dict[str, any]:
        """Return experiment metadata for ExperimentData."""
        metadata = super()._metadata()
        metadata["stark_freq_offset"] = self.experiment_options.stark_freq_offset

        return metadata



# This code is part of Qiskit.
#
# (C) Copyright IBM 2021-2022.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

"""Analysis result class."""

import copy
import logging
import math
import uuid
import traceback
from typing import Optional, List, Union, Dict, Any

import uncertainties

from qiskit_ibm_experiment import IBMExperimentService, AnalysisResultData
from qiskit_ibm_experiment import ResultQuality
from qiskit.exceptions import QiskitError

from qiskit_experiments.framework.json import (
    ExperimentEncoder,
    ExperimentDecoder,
    _serialize_safe_float,
)

from qiskit_experiments.database_service.device_component import DeviceComponent, to_component
from qiskit_experiments.database_service.exceptions import ExperimentDataError
from qiskit_experiments.framework.package_deps import qiskit_version

LOG = logging.getLogger(__name__)


class AnalysisResult:
    """Class representing an analysis result for an experiment.

    Analysis results can also be stored using the experiments service.

    The field ``db_data`` is a dataclass (`ExperimentDataclass`) containing
    all the data that can be stored with the service and loaded from it, and
    as such is subject to strict conventions.

    Other data fields can be added and used freely, but they won't be saved
    to the database.

    Note that the ``result_data`` field of the dataclass is by itself a dictionary
    capable of holding arbitrary values (in a dictionary indexed by a string).

    The data fields in the ``db_data`` dataclass are:

    * ``experiment_id``: ``str``
    * ``result_id``: ``str``
    * ``result_type``: ``str``
    * ``device_components``: ``List[str]``
    * ``quality``: ``str``
    * ``verified``: ``bool``
    * ``tags``: ``List[str]``
    * ``backend_name``: ``str``
    * ``chisq``: ``float``
    * ``result_data``: ``Dict[str]``

    Analysis data that does not fit into the other fields should be added to
    the ``result_data`` dict, e.g. curve parameters in experiments doing a curve fit.
    """

    version = 1
    _data_version = 1

    _json_encoder = ExperimentEncoder
    _json_decoder = ExperimentDecoder

    _extra_data = {}

    RESULT_QUALITY_TO_TEXT = {
        ResultQuality.GOOD: "good",
        ResultQuality.BAD: "bad",
        ResultQuality.UNKNOWN: "unknown",
    }

    def __init__(
        self,
        name: str = None,
        value: Any = None,
        device_components: List[Union[DeviceComponent, str]] = None,
        experiment_id: str = None,
        result_id: Optional[str] = None,
        chisq: Optional[float] = None,
        quality: Optional[str] = RESULT_QUALITY_TO_TEXT[ResultQuality.UNKNOWN],
        extra: Optional[Dict[str, Any]] = None,
        verified: bool = False,
        tags: Optional[List[str]] = None,
        service: Optional[IBMExperimentService] = None,
        source: Optional[Dict[str, str]] = None,
    ) -> "AnalysisResult":
        """AnalysisResult constructor.

        Args:
            name: analysis result name.
            value: main analysis result value.
            device_components: Target device components this analysis is for.
            experiment_id: ID of the experiment.
            result_id: Result ID. If ``None``, one is generated.
            chisq: Reduced chi squared of the fit.
            quality: Quality of the analysis. Refer to the experiment service
                provider for valid values.
            extra: Dictionary of extra analysis result data
            verified: Whether the result quality has been verified.
            tags: Tags for this analysis result.
            service: Experiment service to be used to store result in database.
            source: Class and Qiskit version information when loading from an
                experiment service.
        Returns:
            The AnalysisResult object.
        """
        # Data to be stored in DB.
        self._db_data = AnalysisResultData(
            experiment_id=experiment_id,
            result_id=result_id or str(uuid.uuid4()),
            result_type=name,
            chisq=chisq,
            quality=quality,
            verified=verified,
            tags=tags or [],
        )
        self._db_data.result_data = self.format_result_data(value, extra, chisq, source)
        if device_components is not None:
            for comp in device_components:
                if isinstance(comp, str):
                    comp = to_component(comp)
                self._db_data.device_components.append(comp)

        self._service = service
        self._created_in_db = False
        self._auto_save = False
        if self._service:
            try:
                self.auto_save = self._service.preferences["auto_save"]
            except AttributeError:
                pass

    def set_data(self, data: AnalysisResultData):
        """Sets the analysis data stored in the class"""
        self._db_data = data
        new_device_components = [to_component(comp) for comp in self._db_data.device_components]
        self._db_data.device_components = new_device_components
        self._db_data.quality = self.RESULT_QUALITY_TO_TEXT.get(self._db_data.quality, "unknown")

    @classmethod
    def default_source(cls) -> Dict[str, str]:
        """The default source dictionary to generate"""
        return {
            "class": f"{cls.__module__}.{cls.__name__}",
            "data_version": cls._data_version,
            "qiskit_version": qiskit_version(),
        }

    @staticmethod
    def format_result_data(value, extra, chisq, source):
        """Formats the result data from the given arguments"""
        if source is None:
            source = AnalysisResult.default_source()
        result_data = {
            "_value": copy.deepcopy(value),
            "_chisq": chisq,
            "_extra": copy.deepcopy(extra or {}),
            "_source": source,
        }

        # Format special DB display fields
        if isinstance(value, uncertainties.UFloat):
            db_value = AnalysisResult._display_format(value.nominal_value)
            if db_value is not None:
                result_data["value"] = db_value
            if isinstance(value.std_dev, (int, float)):
                result_data["variance"] = AnalysisResult._display_format(value.std_dev**2)
            if "unit" in result_data["_extra"]:
                result_data["unit"] = result_data["_extra"]["unit"]
        else:
            db_value = AnalysisResult._display_format(value)
            if db_value is not None:
                result_data["value"] = db_value
        return result_data

    @classmethod
    def load(cls, result_id: str, service: IBMExperimentService) -> "AnalysisResult":
        """Load a saved analysis result from a database service.

        Args:
            result_id: Analysis result ID.
            service: the database service.

        Returns:
            The loaded analysis result.
        """
        # Load data from the service
        data = service.analysis_result(result_id, json_decoder=cls._json_decoder)
        result = cls()
        result.set_data(data)
        result._created_in_db = True
        result.service = service
        return result

    def save(self, suppress_errors: bool = True) -> None:
        """Save this analysis result in the database.

        Args:
            suppress_errors: should the method catch exceptions (true) or
                pass them on, potentially aborting the experiment (false).
        Raises:
            ExperimentDataError: If the analysis result contains invalid data.
            QiskitError: If the save to the database failed.
        """
        if not self.service:
            LOG.warning(
                "Analysis result cannot be saved because no experiment service is available."
            )
            return
        try:
            self.service.create_or_update_analysis_result(
                self._db_data, json_encoder=self._json_encoder, create=not self._created_in_db
            )
            self._created_in_db = True
        except Exception as ex:  # pylint: disable=broad-except
            # Don't automatically fail the experiment just because its data cannot be saved.
            LOG.error("Unable to save the experiment data: %s", traceback.format_exc())
            if not suppress_errors:
                raise QiskitError(f"Analysis result save failed\nError Message:\n{str(ex)}") from ex

    def copy(self) -> "AnalysisResult":
        """Return a copy of the result with a new result ID"""
        new_instance = AnalysisResult(
            name=self.name,
            value=self.value,
            device_components=self.device_components,
            experiment_id=self.experiment_id,
        )
        new_instance._db_data = self._db_data.copy()
        new_instance._db_data.result_id = str(uuid.uuid4())
        return new_instance

    @property
    def name(self) -> str:
        """Return analysis result name.

        Returns:
            Analysis result name.
        """
        return self._db_data.result_type

    @property
    def value(self) -> Any:
        """Return analysis result value.

        Returns:
            Analysis result value.
        """
        return self._db_data.result_data["_value"]

    @value.setter
    def value(self, new_value: Any) -> None:
        """Set the analysis result value."""
        self._db_data.result_data["_value"] = new_value
        if self.auto_save:
            self.save()

    @property
    def extra(self) -> Dict[str, Any]:
        """Return extra analysis result data.

        Returns:
            Additional analysis result data.
        """
        return self._db_data.result_data["_extra"]

    @extra.setter
    def extra(self, new_value: Dict[str, Any]) -> None:
        """Set the analysis result value."""
        if not isinstance(new_value, dict):
            raise ExperimentDataError(f"The `extra` field of {type(self).__name__} must be a dict.")
        self._db_data.result_data["_extra"] = new_value
        if self.auto_save:
            self.save()

    @property
    def device_components(self) -> List[DeviceComponent]:
        """Return target device components for this analysis result.

        Returns:
            Target device components.
        """
        return self._db_data.device_components

    @device_components.setter
    def device_components(self, components: List[Union[DeviceComponent, str]]):
        """Set the device components"""
        self._db_data.device_components = []
        for comp in components:
            if isinstance(comp, str):
                comp = to_component(comp)
            self._db_data.device_components.append(comp)

    @property
    def result_id(self) -> str:
        """Return analysis result ID.

        Returns:
            ID for this analysis result.
        """
        return self._db_data.result_id

    @property
    def experiment_id(self) -> str:
        """Return the ID of the experiment associated with this analysis result.

        Returns:
            ID of experiment associated with this analysis result.
        """
        return self._db_data.experiment_id

    @experiment_id.setter
    def experiment_id(self, new_id: str) -> None:
        """Sets the experiment id"""
        self._db_data.experiment_id = new_id

    @property
    def chisq(self) -> Optional[float]:
        """Return the reduced χ² of this analysis."""
        return self._db_data.chisq

    @chisq.setter
    def chisq(self, new_chisq: float) -> None:
        """Set the reduced χ² of this analysis."""
        self._db_data.chisq = new_chisq
        if self.auto_save:
            self.save()

    @property
    def quality(self) -> str:
        """Return the quality of this analysis.

        Returns:
            Quality of this analysis.
        """
        return self._db_data.quality

    @quality.setter
    def quality(self, new_quality: str) -> None:
        """Set the quality of this analysis.

        Args:
            new_quality: New analysis quality.
        """
        self._db_data.quality = new_quality
        if self.auto_save:
            self.save()

    @property
    def verified(self) -> bool:
        """Return the verified flag.

        The ``verified`` flag is intended to indicate whether the quality
        value has been verified by a human.

        Returns:
            Whether the quality has been verified.
        """
        return self._db_data.verified

    @verified.setter
    def verified(self, verified: bool) -> None:
        """Set the verified flag.

        Args:
            verified: Whether the quality is verified.
        """
        self._db_data.verified = verified
        if self.auto_save:
            self.save()

    @property
    def tags(self):
        """Return tags associated with this result."""
        return self._db_data.tags

    @tags.setter
    def tags(self, new_tags: List[str]) -> None:
        """Set tags for this result."""
        if not isinstance(new_tags, list):
            raise ExperimentDataError(f"The `tags` field of {type(self).__name__} must be a list.")
        self._db_data.tags = new_tags
        if self.auto_save:
            self.save()

    @property
    def service(self) -> Optional[IBMExperimentService]:
        """Return the database service.

        Returns:
            Service that can be used to store this analysis result in a database.
            ``None`` if not available.
        """
        return self._service

    @service.setter
    def service(self, service: IBMExperimentService) -> None:
        """Set the service to be used for storing result data in a database.

        Args:
            service: Service to be used.

        Raises:
            ExperimentDataError: If an experiment service is already being used.
        """
        if self._service:
            raise ExperimentDataError("An experiment service is already being used.")
        self._service = service

    @property
    def source(self) -> Dict:
        """Return the class name and version."""
        if "_source" in self._db_data.result_data:
            return self._db_data.result_data["_source"]
        return None

    @property
    def auto_save(self) -> bool:
        """Return current auto-save option.

        Returns:
            Whether changes will be automatically saved.
        """
        return self._auto_save

    @auto_save.setter
    def auto_save(self, save_val: bool) -> None:
        """Set auto save preference.

        Args:
            save_val: Whether to do auto-save.
        """
        if save_val and not self._auto_save:
            self.save()
        self._auto_save = save_val

    @staticmethod
    def _display_format(value):
        """Format values for supported types for display in database service"""
        if value is None or isinstance(value, (int, bool, str)):
            # Pass supported value types directly
            return value
        if isinstance(value, float):
            # Safe handling on NaN float values that serialize to invalid JSON
            if math.isfinite(value):
                return value
            else:
                return _serialize_safe_float(value)["__value__"]
        if isinstance(value, complex):
            # Convert complex floats to strings for display
            return f"{value}"
        # For all other value types that cannot be natively displayed
        # we return the class name
        return f"({type(value).__name__})"

    def __str__(self):
        ret = f"{type(self).__name__}"
        ret += f"\n- name: {self.name}"
        ret += f"\n- value: {str(self.value)}"
        if self.chisq is not None:
            ret += f"\n- χ²: {str(self.chisq)}"
        if self.quality is not None:
            ret += f"\n- quality: {self.quality}"
        if self.extra:
            ret += f"\n- extra: <{len(self.extra)} items>"
        ret += f"\n- device_components: {[str(i) for i in self.device_components]}"
        ret += f"\n- verified: {self.verified}"
        return ret

    def __repr__(self):
        out = f"{type(self).__name__}("
        out += f"name={self.name}"
        out += f", value={repr(self.value)}"
        out += f", device_components={repr(self.device_components)}"
        out += f", experiment_id={self.experiment_id}"
        out += f", result_id={self.result_id}"
        out += f", chisq={self.chisq}"
        out += f", quality={self.quality}"
        out += f", verified={self.verified}"
        out += f", extra={repr(self.extra)}"
        out += f", tags={self.tags}"
        out += f", experiment id={repr(self.experiment_id)}"
        for key, val in self._extra_data.items():
            out += f", {key}={repr(val)}"
        out += ")"
        return out

    def __json_encode__(self):
        return {
            "name": self._db_data.result_type,
            "value": self._db_data.result_data.get("_value", None),
            "device_components": self._db_data.device_components,
            "experiment_id": self._db_data.experiment_id,
            "result_id": self._db_data.result_id,
            "chisq": self._db_data.chisq,
            "quality": self._db_data.quality,
            "extra": self.extra,
            "verified": self._db_data.verified,
            "tags": self._db_data.tags,
            "source": self.source,
        }



# This code is part of Qiskit.
#
# (C) Copyright IBM 2021.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

"""Spectroscopy for the e-f transition."""

from typing import Iterable, Optional, Sequence
from qiskit import QuantumCircuit
from qiskit.providers import Backend
from qiskit.circuit import Gate

from qiskit_experiments.curve_analysis import ParameterRepr
from qiskit_experiments.library.characterization.qubit_spectroscopy import QubitSpectroscopy


class EFSpectroscopy(QubitSpectroscopy):
    """A spectroscopy experiment to obtain a frequency sweep of the qubit's e-f transition.

    # section: overview
        The circuits produced by spectroscopy, i.e.

        .. parsed-literal::

                       ┌───┐┌────────────┐ ░ ┌─┐
                  q_0: ┤ X ├┤ Spec(freq) ├─░─┤M├
                       └───┘└────────────┘ ░ └╥┘
            measure: 1/═══════════════════════╩═
                                              0
    # section: example
        .. jupyter-execute::
            :hide-code:

            # backend
            from qiskit_experiments.test.pulse_backend import SingleTransmonTestBackend
            backend = SingleTransmonTestBackend(5.2e9,-.25e9, 1e9, 0.8e9, noise=False, seed=100)

        .. jupyter-execute::

            import numpy as np
            from qiskit_experiments.library.characterization import EFSpectroscopy

            qubit = 0
            freq01_estimate = backend.defaults().qubit_freq_est[qubit]
            frequencies = np.linspace(freq01_estimate-15e6, freq01_estimate+15e6, 51)

            exp = EFSpectroscopy(physical_qubits = (0,),
                                     frequencies = frequencies,
                                     backend = backend,
                                    )
            exp.set_experiment_options(amp=0.005)

            exp_data = exp.run().block_for_results()
            display(exp_data.figure(0))
            exp_data.analysis_results(dataframe=True)
    """

    def __init__(
        self,
        physical_qubits: Sequence[int],
        frequencies: Iterable[float],
        backend: Optional[Backend] = None,
        absolute: bool = True,
    ):
        super().__init__(physical_qubits, frequencies, backend=backend, absolute=absolute)
        self.analysis.set_options(result_parameters=[ParameterRepr("freq", "f12")])

    def _template_circuit(self, freq_param) -> QuantumCircuit:
        """Return the template quantum circuit."""
        circuit = QuantumCircuit(1)
        circuit.x(0)
        circuit.append(Gate(name=self.__spec_gate_name__, num_qubits=1, params=[freq_param]), (0,))
        circuit.measure_active()

        return circuit



# This code is part of Qiskit.
#
# (C) Copyright IBM 2022.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.
"""
T2 Hahn echo Analysis class.
"""
from typing import Union

import numpy as np

import qiskit_experiments.curve_analysis as curve
from qiskit_experiments.data_processing import DataProcessor, Probability

from qiskit_experiments.framework import Options


class T2HahnAnalysis(curve.DecayAnalysis):
    """A class to analyze T2Hahn experiments."""

    @classmethod
    def _default_options(cls) -> Options:
        """Default analysis options."""
        options = super()._default_options()
        options.plotter.set_figure_options(
            xlabel="Delay",
            ylabel="P(0)",
            xval_unit="s",
        )
        options.data_processor = DataProcessor(
            input_key="counts", data_actions=[Probability(outcome="0")]
        )
        options.bounds = {
            "amp": (0.0, 1.0),
            "tau": (0.0, np.inf),
            "base": (0.0, 1.0),
        }
        options.result_parameters = [curve.ParameterRepr("tau", "T2", "s")]

        return options

    def _evaluate_quality(self, fit_data: curve.CurveFitResult) -> Union[str, None]:
        """Algorithmic criteria for whether the fit is good or bad.

        A good fit has:
            - a reduced chi-squared lower than three and greater than zero
            - absolute amp is within [0.4, 0.6]
            - base is less is within [0.4, 0.6]
            - amp error is less than 0.1
            - tau error is less than its value
            - base error is less than 0.1
        """
        amp = fit_data.ufloat_params["amp"]
        tau = fit_data.ufloat_params["tau"]
        base = fit_data.ufloat_params["base"]

        criteria = [
            0 < fit_data.reduced_chisq < 3,
            abs(amp.nominal_value - 0.5) < 0.1,
            abs(base.nominal_value - 0.5) < 0.1,
            curve.utils.is_error_not_significant(amp, absolute=0.1),
            curve.utils.is_error_not_significant(tau),
            curve.utils.is_error_not_significant(base, absolute=0.1),
        ]

        if all(criteria):
            return "good"

        return "bad"



# This code is part of Qiskit.
#
# (C) Copyright IBM 2021.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

"""Analysis Classes"""

from .drag_analysis import DragCalAnalysis
from .fine_amplitude_analysis import FineAmplitudeAnalysis
from .ramsey_xy_analysis import RamseyXYAnalysis
from .t2ramsey_analysis import T2RamseyAnalysis
from .t2hahn_analysis import T2HahnAnalysis
from .t1_analysis import T1Analysis, T1KerneledAnalysis
from .tphi_analysis import TphiAnalysis
from .cr_hamiltonian_analysis import CrossResonanceHamiltonianAnalysis
from .readout_angle_analysis import ReadoutAngleAnalysis
from .local_readout_error_analysis import LocalReadoutErrorAnalysis
from .correlated_readout_error_analysis import CorrelatedReadoutErrorAnalysis
from .resonator_spectroscopy_analysis import ResonatorSpectroscopyAnalysis
from .zz_ramsey_analysis import ZZRamseyAnalysis
from .multi_state_discrimination_analysis import MultiStateDiscriminationAnalysis



# This code is part of Qiskit.
#
# (C) Copyright IBM 2021.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.
"""
Interleaved RB analysis class.
"""
from typing import List, Union

import lmfit
import numpy as np

import qiskit_experiments.curve_analysis as curve
from qiskit_experiments.framework import AnalysisResultData, ExperimentData


class InterleavedRBAnalysis(curve.CurveAnalysis):
    r"""A class to analyze interleaved randomized benchmarking experiment.

    # section: overview
        This analysis takes only two series for standard and interleaved RB curve fitting.
        From the fit :math:`\alpha` and :math:`\alpha_c` value this analysis estimates
        the error per Clifford (EPC) of the interleaved gate.

        The EPC estimate is obtained using the equation

        .. math::

            r_{\mathcal{C}}^{\text{est}} =
                \frac{\left(d-1\right)\left(1-\alpha_{\overline{\mathcal{C}}}/\alpha\right)}{d}

        The systematic error bounds are given by

        .. math::

            E = \min\left\{
                \begin{array}{c}
                    \frac{\left(d-1\right)\left[\left|\alpha-\alpha_{\overline{\mathcal{C}}}\right|
                    +\left(1-\alpha\right)\right]}{d} \\
                    \frac{2\left(d^{2}-1\right)\left(1-\alpha\right)}
                    {\alpha d^{2}}+\frac{4\sqrt{1-\alpha}\sqrt{d^{2}-1}}{\alpha}
                \end{array}
            \right.

        See Ref. [1] for more details.

    # section: fit_model
        The fit is based on the following decay functions:

        Fit model for standard RB

        .. math::

            F(x) = a \alpha^{x} + b

        Fit model for interleaved RB

        .. math::

            F(x) = a (\alpha_c \alpha)^{x_2} + b

    # section: fit_parameters
        defpar a:
            desc: Height of decay curve.
            init_guess: Determined by :math:`1 - b`.
            bounds: [0, 1]
        defpar b:
            desc: Base line.
            init_guess: Determined by :math:`(1/2)^n` where :math:`n` is number of qubit.
            bounds: [0, 1]
        defpar \alpha:
            desc: Depolarizing parameter.
            init_guess: Determined by :meth:`.rb_decay` with standard RB curve.
            bounds: [0, 1]
        defpar \alpha_c:
            desc: Ratio of the depolarizing parameter of interleaved RB to standard RB.
            init_guess: Determined by alpha of interleaved RB curve divided by one of
                standard RB curve. Both alpha values are estimated by :meth:`.rb_decay`.
            bounds: [0, 1]

    # section: reference
        .. ref_arxiv:: 1 1203.4550

    """

    def __init__(self):
        super().__init__(
            models=[
                lmfit.models.ExpressionModel(
                    expr="a * alpha ** x + b",
                    name="standard",
                ),
                lmfit.models.ExpressionModel(
                    expr="a * (alpha_c * alpha) ** x + b",
                    name="interleaved",
                ),
            ]
        )
        self._num_qubits = None

    @classmethod
    def _default_options(cls):
        """Default analysis options."""
        default_options = super()._default_options()
        default_options.data_subfit_map = {
            "standard": {"interleaved": False},
            "interleaved": {"interleaved": True},
        }
        default_options.result_parameters = ["alpha", "alpha_c"]
        default_options.average_method = "sample"
        return default_options

    def _generate_fit_guesses(
        self,
        user_opt: curve.FitOptions,
        curve_data: curve.ScatterTable,
    ) -> Union[curve.FitOptions, List[curve.FitOptions]]:
        """Create algorithmic initial fit guess from analysis options and curve data.

        Args:
            user_opt: Fit options filled with user provided guess and bounds.
            curve_data: Formatted data collection to fit.

        Returns:
            List of fit options that are passed to the fitter function.
        """
        user_opt.bounds.set_if_empty(
            a=(0, 1),
            alpha=(0, 1),
            alpha_c=(0, 1),
            b=(0, 1),
        )

        b_guess = 1 / 2**self._num_qubits

        # for standard RB curve
        std_curve = curve_data.filter(series="standard")
        alpha_std = curve.guess.rb_decay(std_curve.x, std_curve.y, b=b_guess)
        a_std = (std_curve.y[0] - b_guess) / (alpha_std ** std_curve.x[0])

        # for interleaved RB curve
        int_curve = curve_data.filter(series="interleaved")
        alpha_int = curve.guess.rb_decay(int_curve.x, int_curve.y, b=b_guess)
        a_int = (int_curve.y[0] - b_guess) / (alpha_int ** int_curve.x[0])

        alpha_c = min(alpha_int / alpha_std, 1.0)

        user_opt.p0.set_if_empty(
            b=b_guess,
            a=np.mean([a_std, a_int]),
            alpha=alpha_std,
            alpha_c=alpha_c,
        )

        return user_opt

    def _create_analysis_results(
        self,
        fit_data: curve.CurveFitResult,
        quality: str,
        **metadata,
    ) -> List[AnalysisResultData]:
        """Create analysis results for important fit parameters.

        Args:
            fit_data: Fit outcome.
            quality: Quality of fit outcome.

        Returns:
            List of analysis result data.
        """
        outcomes = super()._create_analysis_results(fit_data, quality, **metadata)

        nrb = 2**self._num_qubits
        scale = (nrb - 1) / nrb

        alpha = fit_data.ufloat_params["alpha"]
        alpha_c = fit_data.ufloat_params["alpha_c"]

        # Calculate epc_est (=r_c^est) - Eq. (4):
        epc = scale * (1 - alpha_c)

        # Calculate the systematic error bounds - Eq. (5):
        systematic_err_1 = scale * (abs(alpha.n - alpha_c.n) + (1 - alpha.n))
        systematic_err_2 = (
            2 * (nrb * nrb - 1) * (1 - alpha.n) / (alpha.n * nrb * nrb)
            + 4 * (np.sqrt(1 - alpha.n)) * (np.sqrt(nrb * nrb - 1)) / alpha.n
        )

        systematic_err = min(systematic_err_1, systematic_err_2)
        systematic_err_l = epc.n - systematic_err
        systematic_err_r = epc.n + systematic_err

        outcomes.append(
            AnalysisResultData(
                name="EPC",
                value=epc,
                chisq=fit_data.reduced_chisq,
                quality=quality,
                extra={
                    "EPC_systematic_err": systematic_err,
                    "EPC_systematic_bounds": [max(systematic_err_l, 0), systematic_err_r],
                    **metadata,
                },
            )
        )

        return outcomes

    def _initialize(
        self,
        experiment_data: ExperimentData,
    ):
        """Initialize curve analysis with experiment data.

        This method is called ahead of other processing.

        Args:
            experiment_data: Experiment data to analyze.
        """
        super()._initialize(experiment_data)

        # Get qubit number
        self._num_qubits = len(experiment_data.metadata["physical_qubits"])



# This code is part of Qiskit.
#
# (C) Copyright IBM 2021.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

"""Probability and phase functions for the mock IQ backend."""

from abc import abstractmethod
from typing import Any, Dict, List, Optional, Tuple
import warnings
import numpy as np
from qiskit import QuantumCircuit
from qiskit.exceptions import QiskitError
from qiskit_aer import AerSimulator
from qiskit_experiments.framework import BaseExperiment

# Define an IQ point typing class.
IQPoint = Tuple[float, float]


class MockIQExperimentHelper:
    """Abstract class for the MockIQ helper classes.

    Different tests will use experiment specific helper classes which define the pattern
    of the IQ data that is then analyzed.
    """

    def __init__(
        self,
        iq_cluster_centers: Optional[List[Tuple[IQPoint, IQPoint]]] = None,
        iq_cluster_width: Optional[List[float]] = None,
    ):
        """Create a MockIQBackend helper object to define how the backend functions.

        :attr:`iq_cluster_centers` and :attr:`iq_cluster_width` define the base IQ
        cluster centers and standard deviations for each qubit in a
        :class:`MockIQBackend` instance. These are used by :meth:`iq_clusters` by
        default. Subclasses can override :meth:`iq_clusters` to return a modified
        version of :attr:`iq_cluster_centers` and :attr:`iq_cluster_width`.
        :attr:`iq_cluster_centers` is a list of tuples. For a given qubit ``i_qbt`` and
        computational state ``i_state`` (either `0` or `1`), the centers of the IQ
        clusters are found by indexing :attr:`iq_cluster_centers` as follows:

        .. code-block:: python

            iq_center = helper.iq_cluster_centers[i_qbt][i_state]
            center_inphase = iq_center[0]
            center_quadrature = iq_center[1]

        :attr:`iq_cluster_width` is indexed similarly except that there is only one width
        per qubit: i.e., the standard deviation of the IQ cluster for qubit ``i_qbt`` is

        .. code-block:: python

            iq_width = helper.iq_cluster_width[i_qbt]

        Subclasses must call ``super().__init__(iq_cluster_centers,iq_cluster_width)`` so that these
        properties are stored appropriately.

        Args:
            iq_cluster_centers: A list of tuples containing the clusters' centers in the IQ plane. There
                are different centers for different logical values of the qubit. Defaults to a single
                qubit with clusters in quadrants 1 and 3.
            iq_cluster_width: A list of standard deviation values for the sampling of each qubit.
                Defaults to widths of 1.0 for each qubit in :attr:`iq_cluster_centers`.

        """
        self._iq_cluster_centers = (
            iq_cluster_centers if iq_cluster_centers is not None else [((-1.0, -1.0), (1.0, 1.0))]
        )
        self._iq_cluster_width = (
            iq_cluster_width
            if iq_cluster_width is not None
            else [1.0] * len(self._iq_cluster_centers)
        )

    @property
    def iq_cluster_centers(self) -> List[Tuple[IQPoint, IQPoint]]:
        """The base cluster centers in the IQ plane."""
        return self._iq_cluster_centers

    @iq_cluster_centers.setter
    def iq_cluster_centers(self, iq_cluster_centers: List[Tuple[IQPoint, IQPoint]]):
        """Set the base cluster centers in the IQ plane."""
        self._iq_cluster_centers = iq_cluster_centers

    @property
    def iq_cluster_width(self) -> List[float]:
        """The base cluster widths in the IQ plane."""
        return self._iq_cluster_width

    @iq_cluster_width.setter
    def iq_cluster_width(self, iq_cluster_width: List[float]):
        """Set the base cluster widths."""
        self._iq_cluster_width = iq_cluster_width

    @abstractmethod
    def compute_probabilities(self, circuits: List[QuantumCircuit]) -> List[Dict[str, Any]]:
        """
        A function provided by the user which is used to determine the probability of each output of the
        circuit. The function returns a list of dictionaries, each containing output binary strings and
        their probabilities.

        Examples:

        **1 qubit circuit - excited state**

        In this experiment, we want to bring a qubit to its excited state and measure it.
        The circuit:

        .. parsed-literal::

                 ┌───┐┌─┐
            q:   ┤ X ├┤M├
                 └───┘└╥┘
            c: 1/══════╩═
                        0

        The function that calculates the probability for this circuit doesn't need any
        calculation parameters:

        .. code-block::

            @staticmethod
            def compute_probabilities(self, circuits: List[QuantumCircuit])
                -> List[Dict[str, float]]:

                output_dict_list = []
                for circuit in circuits:
                    probability_output_dict = {"1": 1.0, "0": 0.0}
                    output_dict_list.append(probability_output_dict)
                return output_dict_list

        **3 qubit circuit**

        In this experiment, we prepare a Bell state with the first and second qubit.
        In addition, we will bring the third qubit to its excited state.
        The circuit:

        .. parsed-literal::

                    ┌───┐     ┌─┐
            q_0:    ┤ H ├──■──┤M├───
                    └───┘┌─┴─┐└╥┘┌─┐
            q_1:    ─────┤ X ├─╫─┤M├
                    ┌───┐└┬─┬┘ ║ └╥┘
            q_2:    ┤ X ├─┤M├──╫──╫─
                    └───┘ └╥┘  ║  ║
            c:    3/═══════╩═══╩══╩═
                        2   0  1

        When an output string isn't in the probability dictionary, the backend will
        assume its probability is 0.

        .. code-block::

            @staticmethod
            def compute_probabilities(self, circuits: List[QuantumCircuit])
                -> List[Dict[str, float]]:

                output_dict_list = []
                for circuit in circuits:
                    probability_output_dict = {}
                    probability_output_dict["001"] = 0.5
                    probability_output_dict["111"] = 0.5
                    output_dict_list.append(probability_output_dict)
                return output_dict_list
        """

    # pylint: disable=unused-argument
    def iq_phase(self, circuits: List[QuantumCircuit]) -> List[float]:
        """Sub-classes can override this method to introduce a phase in the IQ plane.

        This is needed, to test the resonator spectroscopy where the point in the IQ
        plane has a frequency-dependent phase rotation.
        """
        return [0.0] * len(circuits)

    def iq_clusters(
        self,
        circuits: List[QuantumCircuit],
    ) -> List[Tuple[List[Tuple[IQPoint, IQPoint]], List[float]]]:
        """Returns circuit-specific IQ cluster centers and widths in the IQ plane.

        Subclasses can override this function to modify the centers and widths of IQ clusters based on
        the circuits being simulated by a :class:`MockIQBackend`. The base centers and widths are
        stored internally within the helper object, and can be set in :meth:`__init__` or by modifying
        :attr:`iq_cluster_centers` and :attr:`iq_cluster_width`. The default behavior for
        :meth:`iq_clusters` is to return the centers and widths unmodified for each circuit in
        ``circuits``. Subclasses may return different centers and widths based on the circuits provided.

        The returned list contains a tuple per circuit. Each tuple contains the IQ centers and widths in
        the same format as :attr:`iq_cluster_centers` and :attr:`iq_cluster_width`, passed as
        arguments to :meth:`__init__`. The format of the centers and widths lists, in the argument
        list and in the returned tuples, must match the format of :attr:`iq_cluster_centers` and
        :attr:`iq_cluster_width` in :func:`qiskit_experiments.test.MockIQExperimentHelper.__init__`.

        Args:
            circuits: The quantum circuits for which the clusters should be modified.

        Returns:
            List: A list of tuples containing the circuit-specific IQ centers and widths for the
                provided circuits.
        """
        return [(self.iq_cluster_centers, self.iq_cluster_width)] * len(circuits)


class MockIQParallelExperimentHelper(MockIQExperimentHelper):
    """Helper for Parallel experiment."""

    def __init__(
        self,
        exp_list: List[BaseExperiment],
        exp_helper_list: List[MockIQExperimentHelper],
    ):
        """
        Parallel Experiment Helper initializer. The class assumes `exp_helper_list` is ordered to
        match the corresponding experiment in `exp_list`.

        Note that :meth:`__init__` does not have :attr:`iq_cluster_centers` and :attr:`iq_cluster_width`
        as in :func:`MockIQExperimentHelper.__init__`. This is because the centers and widths for
        :class:`MockIQParallelBackend` are stored in multiple experiment helpers in the list
        `exp_helper_list`.

        Args:
            exp_list(List): List of experiments.
            exp_helper_list(List): Ordered list of :class:`.MockIQExperimentHelper` corresponding to the
             experiments in `exp_list`. Nested parallel experiment aren't supported currently.

        Raises:
            ValueError: Raised if the list are empty or if they don't have the same length.
            QiskitError: Raised if `exp_helper_list` contains an object of type
                ``MockIQParallelExperimentHelper``, because the parallel mock backend currently does not
                support parallel sub-experiments.`.

        Examples:

            **Parallel experiment for Resonator Spectroscopy**

            To run a parallel experiment of Resonator Spectroscopy on two qubits we will create two
            instances of `SpectroscopyHelper` objects (for each experiment) and an instance of
            `ParallelExperimentHelper` with them.


            .. code-block::

                iq_cluster_centers = [
                    ((-1.0, 0.0), (1.0, 0.0)),
                    ((0.0, -1.0), (0.0, 1.0)),
                    ((3.0, 0.0), (5.0, 0.0)),
                    ]

                parallel_backend = MockIQParallelBackend(
                    experiment_helper=None,
                    rng_seed=0,
                )
                parallel_backend._configuration.basis_gates = ["x"]
                parallel_backend._configuration.timing_constraints = {"granularity": 16}

                # experiment parameters
                qubit1 = 0
                qubit2 = 1
                freq01 = parallel_backend.defaults().qubit_freq_est[qubit1]
                freq02 = parallel_backend.defaults().qubit_freq_est[qubit2]

                # experiments initialization
                frequencies1 = np.linspace(freq01 - 10.0e6, freq01 + 10.0e6, 23)
                frequencies2 = np.linspace(freq02 - 10.0e6, freq02 + 10.0e6, 21)

                exp_list = [
                    QubitSpectroscopy(qubit1, frequencies1),
                    QubitSpectroscopy(qubit2, frequencies2),
                ]

                exp_helper_list = [
                    SpectroscopyHelper(iq_cluster_centers=iq_cluster_centers,),
                    SpectroscopyHelper(iq_cluster_centers=iq_cluster_centers,),
                ]
                parallel_helper = ParallelExperimentHelper(exp_list, exp_helper_list)

                parallel_backend.experiment_helper = parallel_helper

                # initializing the parallel experiment
                par_experiment = ParallelExperiment(exp_list, backend=parallel_backend)
                par_experiment.set_run_options(meas_level=MeasLevel.KERNELED, meas_return="single")

                par_data = par_experiment.run().block_for_results()
        """
        # Set ParallelExperimentHelper iq_cluster_[centers,widths] to None as exp_helper_list contains
        # the necessary IQ cluster information.
        super().__init__(None, None)

        # check parameters
        self._verify_parameters(exp_list, exp_helper_list)

        self.exp_helper_list = exp_helper_list
        self.exp_list = exp_list

    def compute_probabilities(
        self,
        circuits: List[QuantumCircuit],
    ) -> List[Dict[str, Any]]:
        """
        Run the compute_probabilities for each helper.

        Args:
            circuits: The quantum circuits for which the probabilities should be computed.

        Returns:
            List: A list of dictionaries containing computed probabilities and data for the given
                circuits.
        """
        # checking for legal parameters before computing output.
        self._verify_parameters(self.exp_list, self.exp_helper_list)

        # Splitting the circuit
        parallel_circ_list = self._parallel_exp_circ_splitter(circuits)
        number_of_experiments = len(self.exp_helper_list)
        prob_help_list = [{} for _ in range(number_of_experiments)]

        for idx, (exp_helper, experiment, experiment_circuits) in enumerate(
            zip(self.exp_helper_list, self.exp_list, parallel_circ_list)
        ):
            # Get centers and widths for experiment_circuits and split into centers and widths lists.
            centers_and_widths = exp_helper.iq_clusters(experiment_circuits)
            exp_centers = [c_and_w[0] for c_and_w in centers_and_widths]
            exp_widths = [c_and_w[1] for c_and_w in centers_and_widths]

            prob_help_list[idx] = {
                "physical_qubits": experiment.physical_qubits,
                "prob": exp_helper.compute_probabilities(experiment_circuits),
                "phase": exp_helper.iq_phase(experiment_circuits),
                "centers": exp_centers,
                "widths": exp_widths,
                "num_circuits": len(experiment_circuits),
            }

        return prob_help_list

    def _verify_parameters(
        self,
        exp_list: List[BaseExperiment] = None,
        exp_helper_list: List[MockIQExperimentHelper] = None,
    ):
        """Check parameters before computing probability"""
        if exp_helper_list is None:
            raise ValueError("Please set the experiment helper list.")
        if exp_list is None:
            raise ValueError("Please set the experiment list.")

        number_of_experiments = len(exp_list)
        number_of_helpers = len(exp_helper_list)

        if number_of_experiments == 0:
            raise ValueError("The experiment list cannot be empty.")
        if number_of_helpers == 0:
            raise ValueError("The experiment helper list cannot be empty.")

        if number_of_experiments != number_of_helpers:
            raise ValueError(
                f"The number of helpers {number_of_experiments} and the number of "
                f"experiment {number_of_helpers} don't match."
            )

        for helper in exp_helper_list:
            # checking there is no nested parallel experiment.
            if isinstance(helper, MockIQParallelExperimentHelper):
                raise QiskitError("Nested parallel experiments aren't currently supported.")

    def _parallel_exp_circ_splitter(self, qc_list: List[QuantumCircuit]):
        """
        Splits quantum circuits to their parallel components.
        Args:
            qc_list: The list of quantum circuits the backend gets as input.

        Returns:
            List: A list for each experiment. Each entry is a list of quantum circuits corresponding to
            that experiment.

        Raises:
            QiskitError: If an instruction is applied with qubits that don't belong to the same
            experiment.
            TypeError: The data type provided doesn't match the expected type (`tuple` or `int`).
        """
        exp_circuits_list = [[] for _ in self.exp_list]
        qubits_expid_map = {exp.physical_qubits: i for i, exp in enumerate(self.exp_list)}

        for qc in qc_list:
            # initialize quantum circuit for each experiment for this instance of circuit to fill
            # with instructions.
            for i in range(len(self.exp_list)):
                # we copy the circuit to ensure that the circuit properties (e.g. calibrations and qubit
                # frequencies) are the same in the new circuit.
                empty_qc = qc.copy_empty_like()
                empty_qc.metadata.clear()
                exp_circuits_list[i].append(empty_qc)

            # fixing metadata
            for exp_idx, sub_metadata in zip(
                qc.metadata["composite_index"],
                qc.metadata["composite_metadata"],
            ):
                exp_circuits_list[exp_idx][-1].metadata = sub_metadata.copy()

            # sorting instructions by qubits indexes and inserting them into a circuit of the relevant
            # experiment
            for data in qc.data:
                inst = data.operation
                qarg = data.qubits
                carg = data.clbits
                qubit_indices = set(qc.find_bit(qr).index for qr in qarg)
                for qubits, exp_idx in qubits_expid_map.items():
                    if qubit_indices.issubset(qubits):
                        exp_circuits_list[exp_idx][-1].append(inst, qarg, carg)
                        break
                else:
                    raise QiskitError(
                        "A gate operates on two qubits that don't belong to the same experiment."
                    )

            # deleting empty circuits
            for exp_circuits in exp_circuits_list:
                # 'exp_circuits' is a list of circuits of a specific experiment
                if not exp_circuits[-1].data:
                    exp_circuits.pop()

        return exp_circuits_list


class MockIQDragHelper(MockIQExperimentHelper):
    """Functions needed for test_drag"""

    def __init__(
        self,
        gate_name: str = "Rp",
        ideal_beta: float = 2.0,
        frequency: float = 0.02,
        max_probability: float = 1.0,
        offset_probability: float = 0.0,
        iq_cluster_centers: Optional[List[Tuple[IQPoint, IQPoint]]] = None,
        iq_cluster_width: Optional[List[float]] = None,
    ):
        """
        Args:
            gate_name: name of the gate to count when determining the number of gate repetitions,
            i.e., positive rotation followed by negative rotation, in the circuit.
            ideal_beta: the beta where the minimum of the Drag patterns will be.
            frequency: controls the frequency of the oscillation in the measured Drag pattern.
            max_probability:  a factor to scale the maximum probability of measuring an excited state to
            allow tests to factor in non-ideal situations.
            offset_probability: a constant offset applied to all probabilities to reflect non-ideal
            measurement situations.
            iq_cluster_centers: A list of tuples containing the clusters' centers in the IQ plane. There
            are different centers for different logical values of the qubit.
            iq_cluster_width: A list of standard deviation values for the sampling of each qubit.
        Raises:
            ValueError: If probability value is not valid.
        """
        super().__init__(iq_cluster_centers, iq_cluster_width)
        if max_probability + offset_probability > 1:
            raise ValueError("Probabilities need to be between 0 and 1.")

        self.gate_name = gate_name
        self.ideal_beta = ideal_beta
        self.frequency = frequency
        self.max_probability = max_probability
        self.offset_probability = offset_probability

    def compute_probabilities(self, circuits: List[QuantumCircuit]) -> List[Dict[str, float]]:
        """Returns the probability based on the beta, number of gates, and leakage."""

        gate_name = self.gate_name
        ideal_beta = self.ideal_beta
        freq = self.frequency
        max_prob = self.max_probability
        offset_prob = self.offset_probability

        if max_prob + offset_prob > 1:
            raise ValueError("Probabilities need to be between 0 and 1.")

        output_dict_list = []
        for circuit in circuits:
            probability_output_dict = {}
            n_gates = circuit.count_ops()[gate_name]
            beta = next(iter(circuit.calibrations[gate_name].keys()))[1][0]

            # Dictionary of output string vectors and their probability
            prob = np.sin(2 * np.pi * n_gates * freq * (beta - ideal_beta) / 4) ** 2
            probability_output_dict["1"] = max_prob * prob + offset_prob
            probability_output_dict["0"] = 1 - probability_output_dict["1"]
            output_dict_list.append(probability_output_dict)
        return output_dict_list


class MockIQFineDragHelper(MockIQExperimentHelper):
    """Functions needed for Fine Drag Experiment"""

    def __init__(
        self,
        error: float = 0.03,
        iq_cluster_centers: Optional[List[Tuple[IQPoint, IQPoint]]] = None,
        iq_cluster_width: Optional[List[float]] = None,
    ):
        super().__init__(iq_cluster_centers, iq_cluster_width)
        self.error = error

    def compute_probabilities(self, circuits: List[QuantumCircuit]) -> List[Dict[str, float]]:
        """Returns the probability based on error per gate."""

        error = self.error
        output_dict_list = []
        for circuit in circuits:
            probability_output_dict = {}
            n_gates = circuit.count_ops().get("rz", 0) // 2

            # Dictionary of output string vectors and their probability
            probability_output_dict["1"] = 0.5 * np.sin(n_gates * error) + 0.5
            probability_output_dict["0"] = 1 - probability_output_dict["1"]
            output_dict_list.append(probability_output_dict)
        return output_dict_list


class MockIQRabiHelper(MockIQExperimentHelper):
    """Functions needed for Rabi experiment on mock IQ backend"""

    def __init__(
        self,
        amplitude_to_angle: float = np.pi,
        iq_cluster_centers: Optional[List[Tuple[IQPoint, IQPoint]]] = None,
        iq_cluster_width: Optional[List[float]] = None,
    ):
        """
        Args:
            amplitude_to_angle: maps a pulse amplitude to a rotation angle.
        """
        warnings.warn(
            "MockIQRabiHelper has been deprecated. It will be removed "
            "in Qiskit Experiments 0.5.",
            DeprecationWarning,
            stacklevel=2,
        )

        super().__init__(iq_cluster_centers, iq_cluster_width)
        self.amplitude_to_angle = amplitude_to_angle

    def compute_probabilities(self, circuits: List[QuantumCircuit]) -> List[Dict[str, float]]:
        """Returns the probability based on the rotation angle and amplitude_to_angle."""
        amplitude_to_angle = self.amplitude_to_angle
        output_dict_list = []
        for circuit in circuits:
            probability_output_dict = {}
            amp = next(iter(circuit.calibrations["Rabi"].keys()))[1][0]

            # Dictionary of output string vectors and their probability
            probability_output_dict["1"] = np.sin(amplitude_to_angle * amp) ** 2
            probability_output_dict["0"] = 1 - probability_output_dict["1"]
            output_dict_list.append(probability_output_dict)
        return output_dict_list

    def rabi_rate(self) -> float:
        """Returns the rabi rate."""
        return self.amplitude_to_angle / np.pi


class MockIQFineFreqHelper(MockIQExperimentHelper):
    """Functions needed for Fine Frequency experiment on mock IQ backend"""

    def __init__(
        self,
        sx_duration: float = 160,
        freq_shift: float = 0,
        dt: float = 1e-9,
        iq_cluster_centers: Optional[List[Tuple[IQPoint, IQPoint]]] = None,
        iq_cluster_width: Optional[List[float]] = None,
    ):
        """
        Args:
            sx_duration: duration of the single-qubit sx gate.
            freq_shift: the detuning from the ideal frequency that this mock backend will mimic.
            dt: duration of a sample.
        """
        super().__init__(iq_cluster_centers, iq_cluster_width)
        self.sx_duration = sx_duration
        self.freq_shift = freq_shift
        self.dt = dt

    def compute_probabilities(self, circuits: List[QuantumCircuit]) -> List[Dict[str, float]]:
        """Return the probability of being in the excited state."""
        sx_duration = self.sx_duration
        freq_shift = self.freq_shift
        dt = self.dt
        simulator = AerSimulator(method="automatic")
        output_dict_list = []
        for circuit in circuits:
            probability_output_dict = {}
            delay = None
            for instruction in circuit.data:
                if instruction.operation.name == "delay":
                    delay = instruction.operation.duration

            if delay is None:
                probability_output_dict = {"1": 1, "0": 0}
            else:
                reps = delay // sx_duration

                qc = QuantumCircuit(1)
                qc.sx(0)
                qc.rz(np.pi * reps / 2 + 2 * np.pi * freq_shift * delay * dt, 0)
                qc.sx(0)
                qc.measure_all()

                counts = simulator.run(qc, seed_simulator=1).result().get_counts(0)
                probability_output_dict["1"] = counts.get("1", 0) / sum(counts.values())
                probability_output_dict["0"] = 1 - probability_output_dict["1"]
            output_dict_list.append(probability_output_dict)

        return output_dict_list


class MockIQFineAmpHelper(MockIQExperimentHelper):
    """Functions needed for Fine Amplitude experiment on mock IQ backend"""

    def __init__(
        self,
        angle_error: float = 0,
        angle_per_gate: float = 0,
        gate_name: str = "x",
        iq_cluster_centers: Optional[List[Tuple[IQPoint, IQPoint]]] = None,
        iq_cluster_width: Optional[List[float]] = None,
    ):
        """
        Args:
            angle_error: rotation angle error per gate.
            angle_per_gate: the intended rotation angle per gate.
            gate_name: name of the gate that will be counted to determine the total rotation.
        """
        super().__init__(iq_cluster_centers, iq_cluster_width)
        self.angle_error = angle_error
        self.angle_per_gate = angle_per_gate
        self.gate_name = gate_name

    def compute_probabilities(self, circuits: List[QuantumCircuit]) -> List[Dict[str, float]]:
        """Return the probability of being in the excited state."""
        angle_error = self.angle_error
        angle_per_gate = self.angle_per_gate
        gate_name = self.gate_name
        output_dict_list = []
        for circuit in circuits:
            probability_output_dict = {}
            n_ops = circuit.count_ops().get(gate_name, 0)
            angle = n_ops * (angle_per_gate + angle_error)

            if gate_name != "sx":
                angle += np.pi / 2 * circuit.count_ops().get("sx", 0)

            if gate_name != "x":
                angle += np.pi * circuit.count_ops().get("x", 0)

            # Dictionary of output string vectors and their probability
            probability_output_dict["1"] = np.sin(angle / 2) ** 2
            probability_output_dict["0"] = 1 - probability_output_dict["1"]
            output_dict_list.append(probability_output_dict)

        return output_dict_list


class MockIQRamseyXYHelper(MockIQExperimentHelper):
    """Functions needed for Ramsey XY experiment on mock IQ backend"""

    def __init__(
        self,
        t2ramsey: float = 100e-6,
        freq_shift: float = 0,
        iq_cluster_centers: Optional[List[Tuple[IQPoint, IQPoint]]] = None,
        iq_cluster_width: Optional[List[float]] = None,
    ):
        super().__init__(iq_cluster_centers, iq_cluster_width)
        self.t2ramsey = t2ramsey
        self.freq_shift = freq_shift

    def compute_probabilities(self, circuits: List[QuantumCircuit]) -> List[Dict[str, float]]:
        """Return the probability of being in the excited state."""
        t2ramsey = self.t2ramsey
        freq_shift = self.freq_shift
        output_dict_list = []
        for circuit in circuits:
            probability_output_dict = {}
            series = circuit.metadata["series"]
            delay = circuit.metadata["xval"]

            if series == "X":
                phase_offset = 0.0
            else:
                phase_offset = np.pi / 2

            probability_output_dict["1"] = (
                0.5
                * np.exp(-delay / t2ramsey)
                * np.cos(2 * np.pi * delay * freq_shift - phase_offset)
                + 0.5
            )
            probability_output_dict["0"] = 1 - probability_output_dict["1"]
            output_dict_list.append(probability_output_dict)
        return output_dict_list


class MockIQSpectroscopyHelper(MockIQExperimentHelper):
    """Functions needed for Spectroscopy experiment on mock IQ backend"""

    def __init__(
        self,
        gate_name: str = "Spec",
        freq_offset: float = 0.0,
        line_width: float = 2e6,
        iq_cluster_centers: Optional[List[Tuple[IQPoint, IQPoint]]] = None,
        iq_cluster_width: Optional[List[float]] = None,
    ):
        """
        Args:
            gate_name: the gate name to look for when calculating frequency shift.
            freq_offset: frequency offset from resonance that this mock backend will mimic.
            line_width: line width of the resonance of the spectroscopy signal.
        """
        super().__init__(iq_cluster_centers, iq_cluster_width)
        self.freq_offset = freq_offset
        self.line_width = line_width
        self.gate_name = gate_name

    def compute_probabilities(self, circuits: List[QuantumCircuit]) -> List[Dict[str, float]]:
        """Returns the probability based on the parameters provided."""
        freq_offset = self.freq_offset
        line_width = self.line_width
        output_dict_list = []
        for circuit in circuits:
            probability_output_dict = {}
            if self.gate_name == "measure":
                freq_shift = (
                    next(iter(circuit.calibrations[self.gate_name].values())).blocks[0].frequency
                )
            elif self.gate_name == "Spec":
                freq_shift = next(iter(circuit.calibrations[self.gate_name]))[1][0]
            else:
                raise ValueError(f"The gate name {str(self.gate_name)} isn't supported.")
            delta_freq = freq_shift - freq_offset

            probability_output_dict["1"] = np.abs(1 / (1 + 2.0j * delta_freq / line_width))
            probability_output_dict["0"] = 1 - probability_output_dict["1"]
            output_dict_list.append(probability_output_dict)
        return output_dict_list

    def iq_phase(self, circuits: List[QuantumCircuit]) -> List[float]:
        """Add a phase to the IQ point depending on how far we are from the resonance.
        This will cause the IQ points to rotate around in the IQ plane when we approach the
        resonance, introducing extra complication that the data processor has to
        properly handle.
        """
        delta_freq_list = [0.0] * len(circuits)
        if self.gate_name == "measure":

            for circ_idx, circ in enumerate(circuits):
                freq_shift = next(iter(circ.calibrations["measure"].values())).blocks[0].frequency
                delta_freq_list[circ_idx] = freq_shift - self.freq_offset
        phase = [delta_freq / self.line_width for delta_freq in delta_freq_list]
        return phase


class MockIQReadoutAngleHelper(MockIQExperimentHelper):
    """Functions needed for Readout angle experiment on mock IQ backend"""

    def compute_probabilities(self, circuits: List[QuantumCircuit]) -> List[Dict[str, float]]:
        """Return the probability of being in the excited state."""
        output_dict_list = []
        for circuit in circuits:
            probability_output_dict = {"1": 1 - circuit.metadata["xval"]}
            probability_output_dict["0"] = 1 - probability_output_dict["1"]
            output_dict_list.append(probability_output_dict)

        return output_dict_list


class MockIQHalfAngleHelper(MockIQExperimentHelper):
    """Functions needed for Half Angle experiment on mock IQ backend"""

    def __init__(
        self,
        error: float = 0,
        iq_cluster_centers: Optional[List[Tuple[IQPoint, IQPoint]]] = None,
        iq_cluster_width: Optional[List[float]] = None,
    ):
        super().__init__(iq_cluster_centers, iq_cluster_width)
        self.error = error

    def compute_probabilities(self, circuits: List[QuantumCircuit]) -> List[Dict[str, float]]:
        """Return the probability of being in the excited state."""
        error = self.error
        output_dict_list = []
        for circuit in circuits:
            probability_output_dict = {}
            n_gates = circuit.metadata["xval"]

            # Dictionary of output string vectors and their probability
            probability_output_dict["1"] = (
                0.5 * np.sin((-1) ** (n_gates + 1) * n_gates * error) + 0.5
            )
            probability_output_dict["0"] = 1 - probability_output_dict["1"]
            output_dict_list.append(probability_output_dict)

        return output_dict_list


class MockIQT1Helper(MockIQExperimentHelper):
    """Functions needed for T1 experiment on mock IQ backend"""

    def __init__(
        self,
        t1: float = None,
        iq_cluster_centers: Optional[List[Tuple[IQPoint, IQPoint]]] = None,
        iq_cluster_width: Optional[List[float]] = None,
    ):
        super().__init__(iq_cluster_centers, iq_cluster_width)
        self._t1 = t1 or 90e-6

    def compute_probabilities(self, circuits: List[QuantumCircuit]) -> List[Dict[str, float]]:
        """Return the probability of being in the excited state."""
        output_dict_list = []
        for circuit in circuits:
            probability_output_dict = {}

            # extracting information from the circuit.
            delay = circuit.metadata["xval"]

            # creating a probability dict.
            probability_output_dict["1"] = np.exp(-delay / self._t1)
            probability_output_dict["0"] = 1 - probability_output_dict["1"]
            output_dict_list.append(probability_output_dict)

        return output_dict_list



# This code is part of Qiskit.
#
# (C) Copyright IBM 2021.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

"""
=====================================================================
Tomography Experiments (:mod:`qiskit_experiments.library.tomography`)
=====================================================================

.. currentmodule:: qiskit_experiments.library.tomography


Experiments
===========
.. autosummary::
    :toctree: ../stubs/
    :template: autosummary/experiment.rst

    TomographyExperiment
    StateTomography
    ProcessTomography
    MitigatedStateTomography
    MitigatedProcessTomography


Analysis
========

.. autosummary::
    :toctree: ../stubs/
    :template: autosummary/analysis.rst

    TomographyAnalysis
    StateTomographyAnalysis
    ProcessTomographyAnalysis
    MitigatedTomographyAnalysis

Tomography Fitters
==================

Fitter functions for state reconstruction in tomography analysis

.. autosummary::
    :toctree: ../stubs/

    fitters.linear_inversion
    fitters.cvxpy_gaussian_lstsq
    fitters.cvxpy_linear_lstsq


Basis Classes
=============

Built in tomography basis classes

.. autosummary::
    :toctree: ../stubs/

    basis.PauliMeasurementBasis
    basis.PauliPreparationBasis
    basis.Pauli6PreparationBasis

Custom local tensor product basis classes

.. autosummary::
    :toctree: ../stubs/

    basis.LocalMeasurementBasis
    basis.LocalPreparationBasis

Abstract base classes

.. autosummary::
    :toctree: ../stubs/

    basis.MeasurementBasis
    basis.PreparationBasis

.. warning::
    The API for tomography fitters and bases is still under development so may
    change in a future release.
"""

# Experiment Classes
from .tomography_experiment import TomographyExperiment
from .qst_experiment import StateTomography, StateTomographyAnalysis
from .qpt_experiment import ProcessTomography, ProcessTomographyAnalysis
from .mit_qst_experiment import MitigatedStateTomography
from .mit_qpt_experiment import MitigatedProcessTomography
from .tomography_analysis import TomographyAnalysis
from .mit_tomography_analysis import MitigatedTomographyAnalysis

# Basis Classes
from . import basis
from . import fitters



# This code is part of Qiskit.
#
# (C) Copyright IBM 2021.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.
"""
===========================================================
Data Processing (:mod:`qiskit_experiments.data_processing`)
===========================================================

.. currentmodule:: qiskit_experiments.data_processing

Data processing is the act of taking the data returned by the backend and
converting it into a format that can be analyzed.
It is implemented as a chain of data processing steps that transform various input data,
e.g. IQ data, into a desired format, e.g. population, which can be analyzed.

These data transformations may consist of multiple steps, such as kerneling and discrimination.
Each step is implemented by a :class:`~qiskit_experiments.data_processing.data_action.DataAction`
also called a `node`.

The data processor implements the :meth:`__call__` method. Once initialized, it
can thus be used as a standard python function:

.. code-block:: python

    processor = DataProcessor(input_key="memory", [Node1(), Node2(), ...])
    out_data = processor(in_data)

The data input to the processor is a sequence of dictionaries each representing the result
of a single circuit. The output of the processor is a numpy array whose shape and data type
depend on the combination of the nodes in the data processor.

Uncertainties that arise from quantum measurements or finite sampling can be taken into account
in the nodes: a standard error can be generated in a node and can be propagated
through the subsequent nodes in the data processor.
Correlation between computed values is also considered.


Classes
=======
.. autosummary::
    :toctree: ../stubs/

    DataProcessor
    DataAction
    TrainableDataAction


Data Processing Nodes
=====================
.. autosummary::
    :toctree: ../stubs/

    Probability
    MarginalizeCounts
    ToImag
    ToReal
    SVD
    DiscriminatorNode
    MemoryToCounts
    AverageData
    BasisExpectationValue
    MinMaxNormalize
    ShotOrder
    RestlessNode
    RestlessToCounts
    RestlessToIQ


Discriminators
==============
.. autosummary::
    :toctree: ../stubs/

    BaseDiscriminator
    SkLDA
    SkQDA
"""

from .data_action import DataAction, TrainableDataAction
from .nodes import (
    Probability,
    MarginalizeCounts,
    ToImag,
    ToReal,
    SVD,
    DiscriminatorNode,
    MemoryToCounts,
    AverageData,
    BasisExpectationValue,
    MinMaxNormalize,
    ShotOrder,
    RestlessNode,
    RestlessToCounts,
    RestlessToIQ,
)

from .data_processor import DataProcessor
from .discriminator import BaseDiscriminator
from .sklearn_discriminators import SkLDA, SkQDA



# This code is part of Qiskit.
#
# (C) Copyright IBM 2021.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.
"""
Experiment and analysis config dataclasses.
"""

import dataclasses
from typing import Tuple, Dict, Any

from qiskit.exceptions import QiskitError
from qiskit_experiments.version import __version__


@dataclasses.dataclass(frozen=True)
class ExperimentConfig:
    """Store configuration settings for an Experiment class.

    This stores the current configuration of a :class:`.BaseExperiment` and can be used to
    reconstruct the experiment using either the :meth:`experiment` property if the experiment class
    type is currently stored, or the :meth:`~.BaseExperiment.from_config` class method of the
    appropriate experiment.
    """

    cls: type = None
    args: Tuple[Any] = dataclasses.field(default_factory=tuple)
    kwargs: Dict[str, Any] = dataclasses.field(default_factory=dict)
    experiment_options: Dict[str, Any] = dataclasses.field(default_factory=dict)
    transpile_options: Dict[str, Any] = dataclasses.field(default_factory=dict)
    run_options: Dict[str, Any] = dataclasses.field(default_factory=dict)
    version: str = __version__

    def experiment(self):
        """Return the experiment constructed from this config.

        Returns:
            BaseExperiment: The experiment reconstructed from the config.

        Raises:
            QiskitError: If the experiment class is not stored,
                         was not successful deserialized, or reconstruction
                         of the experiment fails.
        """
        cls = self.cls
        if cls is None:
            raise QiskitError("No experiment class in experiment config")
        if isinstance(cls, dict):
            raise QiskitError(
                "Unable to load experiment class. Try manually loading "
                "experiment using `Experiment.from_config(config)` instead."
            )
        try:
            return cls.from_config(self)
        except Exception as ex:
            msg = "Unable to construct experiments from config."
            if cls.version != __version__:
                msg += (
                    f" Note that config version ({cls.version}) differs from the current"
                    f" qiskit-experiments version ({__version__}). You could try"
                    " installing a compatible qiskit-experiments version."
                )
            raise QiskitError(f"{msg}\nError Message:\n{str(ex)}") from ex


@dataclasses.dataclass(frozen=True)
class AnalysisConfig:
    """Store configuration settings for an Analysis class.

    This stores the current configuration of a :class:`.BaseAnalysis` and can be used to reconstruct
    the analysis class using either the :meth:`analysis` property if the analysis class type is
    currently stored, or the :meth:`~.BaseAnalysis.from_config` class method of the appropriate
    experiment.
    """

    cls: type = None
    args: Tuple[Any] = dataclasses.field(default_factory=tuple)
    kwargs: Dict[str, Any] = dataclasses.field(default_factory=dict)
    options: Dict[str, Any] = dataclasses.field(default_factory=dict)
    version: str = __version__

    def analysis(self):
        """Return the analysis class constructed from this config.

        Returns:
            BaseAnalysis: The analysis reconstructed from the config.

        Raises:
            QiskitError: If the analysis class is not stored,
                         was not successful deserialized, or reconstruction
                         of the analysis class fails.
        """
        cls = self.cls
        if cls is None:
            raise QiskitError("No analysis class in analysis config")
        if isinstance(cls, dict):
            raise QiskitError(
                "Unable to load analysis class. Try manually loading "
                "analysis using `Analysis.from_config(config)` instead."
            )
        try:
            return cls.from_config(self)
        except Exception as ex:
            msg = "Unable to construct analysis from config."
            if cls.version != __version__:
                msg += (
                    f" Note that config version ({cls.version}) differs from the current"
                    f" qiskit-experiments version ({__version__}). You could try"
                    " installing a compatible qiskit-experiments version."
                )
            raise QiskitError(f"{msg}\nError Message:\n{str(ex)}") from ex
